# Jeremie Kalfon
# for BroadInsitute
# in 2019

from __future__ import print_function

import pandas as pd
import numpy as np
from genepy.utils import helper as h
import gzip
import seaborn as sns


def vcf_to_df(
    path,
    additional_cols=[],
    additional_filters=[],
    parse_filter=False,
    drop_null=False,
    force_keep=[],
    **kwargs,
):
    """
    transforms a vcf file into a dataframe file as best as it can

    Args:
    -----
      path: str filepath to the vcf file
      hasfilter: bool whether or not the vcf has a filter column
      samples: list[str] colnames of the sample names.
      additional_cols: list[str] of additional colnames in the vcf already looks for 'DB', 'SOMATIC', 'GERMLINE', "OVERLAP", "IN_PON", "STR", "ReverseComplementedAlleles"

    Returns:
    --------
      a dataframe fo the vcf
      a dict associating each column with its description (gathered from the vcf header)
    """
    uniqueargs = [
        "DB",
        "SOMATIC",
        "GERMLINE",
        "OVERLAP",
        "IN_PON",
        "STR",
        "ReverseComplementedAlleles",
    ] + additional_cols

    filters = [
        "PASS",
        "base_qual",
        "clustered_events",
        "fragment",
        "germline",
        "haplotype",
        "map_qual",
        "multiallelic",
        "panel_of_normals",
        "position",
        "slippage",
        "strand_bias",
        "weak_evidence",
    ] + additional_filters

    FUNCO_DESC = "Functional annotation from the Funcotator tool."

    dropped_cols = []

    def read_comments(f):
        description = {}
        colnames = []
        rows = 0
        for l in f:
            l = l.decode("utf-8") if type(l) is not str else l
            if l.startswith("##"):
                rows += 1
                if "FORMAT" in l[:20]:
                    res = l.split("ID=")[1].split(",")[0]
                    desc = l.split("Description=")[1][:-2]
                    description.update({res: desc})
                if "INFO" in l[:20]:
                    res = l.split("ID=")[1].split(",")[0]
                    if res == "FUNCOTATION":
                        print("parsing funcotator special")
                        for val in l.split("Description=")[1][:-2].split("|"):
                            val = val.split("Funcotation fields are: ")[-1]
                            description.update({val: FUNCO_DESC})
                    else:
                        desc = l.split("Description=")[1][:-2]
                        description.update({res: desc})
            elif l.startswith("#"):
                colnames = l[1:-1].split("\t")
                rows += 1
            else:
                break
        return description, colnames, rows

    if path.endswith(".gz"):
        with gzip.open(path, "r") as f:
            description, colnames, nrows_toskip = read_comments(f)
    else:
        with open(path, "r") as f:
            description, colnames, nrows_toskip = read_comments(f)
    colnames = [i for i in colnames]
    csvkwargs = {
        "sep": "\t",
        "index_col": False,
        "header": None,
        "names": colnames,
        "skiprows": nrows_toskip + kwargs.get("skiprows", 0),
    }
    data = pd.read_csv(path, **{**kwargs, **csvkwargs})
    print(description)
    funco_fields = [k for k, v in description.items() if FUNCO_DESC in v]
    fields = {k: [] for k, _ in description.items()}
    try:
        for j, info in enumerate(data["INFO"].str.split(";").values.tolist()):
            res = {}
            # show speed
            if j % 10_000 == 0:
                print(j, end="\r")
            for annot in info:
                if annot in uniqueargs:
                    res.update({annot: True})
                elif "=" in annot:
                    # taking care of the funcotator special fields
                    if "FUNCOTATION" in annot:
                        # for multi allelic site:
                        annot = annot.replace("FUNCOTATION=", "")[1:-1]
                        res.update({name: [] for name in funco_fields})
                        for site in annot.split("],["):
                            site = (
                                site.replace("_%7C_", " ")
                                .replace("_%20_", " ")
                                .replace("_%2C_", ",")
                                .replace("_%3D_", "=")
                                .split("|")
                            )
                            for i, sub_annot in enumerate(site):
                                res[funco_fields[i]].append(sub_annot)
                        for k in funco_fields:
                            res[k] = ",".join(res[k])
                    else:
                        k, annot = annot.split("=")
                        res.update({k: annot})
                else:
                    raise ValueError("unknown argument: " + annot)
            for k in list(fields.keys()):
                fields[k].append(res.get(k, None))
    except ValueError:
        print(annot)
        raise ValueError("unknown field")

    data = pd.concat(
        [data.drop(columns="INFO"), pd.DataFrame(data=fields, index=data.index)], axis=1
    )
    if drop_null:
        cols_to_drop = []
        for f in funco_fields:
            # drop columns that have the same value across all rows
            uniq = data[f].unique()
            if len(uniq) == 1 and f.lower() not in force_keep:
                cols_to_drop.append(f)
                continue
            elif len(uniq) < 10:
                # checking multi allelic stuff
                multi = []
                for v in uniq:
                    multi += v.split(",")
                if len(set(multi)) == 1 and f.lower() not in force_keep:
                    cols_to_drop.append(f)
        print("dropping uninformative columns:", cols_to_drop)
        data = data.drop(columns=cols_to_drop)
        dropped_cols += cols_to_drop
    data.columns = [i.lower() for i in data.columns]
    samples = [i.lower() for i in colnames[9:]]
    print("\nthe samples are:", samples)
    sorting = data["format"][0].split(":")
    for sample in samples:
        res = data[sample].str.split(":").values.tolist()
        maxcols = max([len(v) for v in res])
        if maxcols - len(sorting) > 0:
            for i in range(maxcols - len(sorting)):
                sorting.append(sorting[-1] + "_" + str(i + 1))
        if len(samples) > 1:
            sorting = [sample + "_" + v for v in sorting]
        data = pd.concat(
            [
                data.drop(columns=sample),
                pd.DataFrame(data=res, columns=sorting, index=data.index),
            ],
            axis=1,
        )

    # subsetting filters
    if parse_filter:
        data[filters] = False
        for f in filters:
            data.loc[data["filter"].str.contains(f), f] = True
        data = data.drop(columns="filter")
        dropped_cols.append("filter")

    # cleaning empty cols
    data = data.drop(columns="format")
    dropped_cols.append("format")

    todrop = []
    for val in ["gt", "ad", "af", "dp", "f1r2", "f2r1", "fad", "sb"]:
        if val in data.columns.tolist():
            todrop.append(val)
    data = data.drop(columns=todrop)

    if drop_null:
        empty = data.columns[data.isna().sum() == len(data)].tolist()
        empty = list(set(empty) - set(force_keep))
        print("dropping empty columns:", empty)
        data = data.drop(columns=empty)
        dropped_cols += empty

    # weird bug sometimes
    if "SB_1" in data.columns.tolist():
        loc = ~data.SB_1.isna()
        data.loc[loc, "PGT"] = data.loc[loc, "SB"]
        data.loc[loc, "SB"] = data.loc[loc, "SB_1_2_3"]
        data = data.drop(columns=["SB_1", "SB_1_2_3"])
        data = data.rename(columns={"SB_1_2": "PS", "SB_1": "PID"})
    else:
        loc = data.SB.isna()
        data.loc[loc, "SB"] = data.loc[loc, "PGT"]
        data.loc[loc, "PGT"] = None
    # sorting out issue with
    return data, description, dropped_cols


def mafToMat(
    maf,
    mode="bool",
    freqcol="tumor_f",
    samplesCol="DepMap_ID",
    mutNameCol="Hugo_Symbol",
    minfreqtocall=0.2,
):
    """
    turns a maf file into a matrix of mutations x samples (works with multiple sample file)

    Args:
    -----
      maf: dataframe of the maf file
      sample_col: str colname for samples
      mode: flag  "bool" to convert the matrix into a boolean (mut/no mut)
                  "float" to keep the allele frequencies as is (0.x)
                  "genotype" to have either 1, 0.5 or 0
      freqcol: str colname where ref/alt frequencies are stored
      mutNameCol: str colname where mutation names are stored, will merge things over that column name

    Returns:
    --------
      the dataframe matrix
    """
    samples = set(maf[samplesCol])
    maf = maf[maf[freqcol] >= minfreqtocall]
    maf = maf.sort_values(by=mutNameCol)
    mut = pd.DataFrame(
        data=np.zeros((len(set(maf[mutNameCol])), 1)),
        columns=["fake"],
        index=set(maf[mutNameCol]),
    ).astype(float)
    for i, val in enumerate(samples):
        h.showcount(i, len(samples))
        if mode == "genotype":
            mut = mut.join(
                maf[maf[samplesCol] == val]
                .set_index(mutNameCol)[freqcol]
                .groupby(mutNameCol)
                .agg("sum")
                .rename(val)
            )
        else:
            mut = mut.join(
                maf[maf[samplesCol] == val]
                .drop_duplicates(mutNameCol)
                .set_index(mutNameCol)[freqcol]
                .rename(val)
            )
    mut = mut.fillna(0).astype(bool if mode == "bool" else float).drop(columns=["fake"])
    if mode == "genotype":
        mut[(mut > 1.3)] = 3
        mut[(mut >= 0.7) & (mut <= 1.3)] = 2
        mut[(mut > 0.3) & (mut < 0.7)] = 1
        mut[mut <= 0.3] = 0
    return mut


def mergeAnnotations(
    firstmaf,
    additionalmaf,
    mutcol="mutation",
    Genome_Change="Genome_Change",
    Start_position="Start_position",
    Chromosome="Chromosome",
    samplename="DepMap_ID",
    useSecondForConflict=True,
    dry_run=False,
):
    """
    merges two maf files, taking carre of duplicate samples and duplicate (works with multiple sample file)

    Args:
    -----
    firstmaf: dataframe the first maf file
    additionalmaf: dataframe the second maf file (need to contain same colnames)
    Genome_Change: str colnames of the Genome_Change column
    Start_position: str colnames of the Start_position column
    Chromosome: str colnames of the Chromosome column
    samplename: str colnames of the samplename column (for multiple samples, even if one, needs to have this column)
    useSecondForConflict: bool if false use the first df as reference else use the second one
    dry_run: if true, will just output conflict regions and not merge the dataframes

    Returns:
    -------
      dataframe of the maf file if not dryrun, else an np array of the merge issues
    """
    mutations = firstmaf.copy()
    mutations["ind"] = mutations[samplename] + "_" + mutations[Genome_Change]
    mutations["loci"] = (
        mutations[samplename]
        + "_"
        + mutations[Chromosome]
        + "_"
        + mutations[Start_position].astype(str)
    )
    additionalmaf["ind"] = (
        additionalmaf[samplename] + "_" + additionalmaf[Genome_Change]
    )
    additionalmaf["loci"] = (
        additionalmaf[samplename]
        + "_"
        + additionalmaf[Chromosome]
        + "_"
        + additionalmaf[Start_position].astype(str)
    )
    inboth = set(additionalmaf["loci"]) & set(mutations["loci"])
    notineach = set(additionalmaf["ind"]) ^ set(mutations["ind"])
    submut = mutations[mutations.loci.isin(inboth) & mutations.ind.isin(notineach)]
    subother = additionalmaf[
        additionalmaf.loci.isin(inboth) & additionalmaf.ind.isin(notineach)
    ]
    issues = None
    if len(submut) > 0:
        print("found " + str(len(submut)) + " nonmatching mutations")
        issues = np.vstack(
            [
                submut.sort_values(by="loci")[Genome_Change].values,
                subother.sort_values(by="loci")[Genome_Change].values,
            ]
        ).T
        if dry_run:
            print(issues)
    if not dry_run:
        if issues is not None:
            if useSecondForConflict:
                mutations = mutations[~mutations.ind.isin(set(submut.ind))]
            else:
                additionalmaf = additionalmaf[
                    ~additionalmaf.ind.isin(set(subother.ind))
                ]
            mutations = mutations.append(
                additionalmaf[
                    additionalmaf["ind"].isin(
                        set(additionalmaf["ind"]) - set(mutations["ind"])
                    )
                ]
            )
        subother = additionalmaf[
            additionalmaf.loci.isin(inboth) & ~additionalmaf.ind.isin(notineach)
        ].set_index("ind")
        mutations = mutations.set_index("ind")
        mutations.loc[subother.index.tolist(), mutcol] = subother[mutcol].tolist()
        return (
            mutations.drop(columns=["loci"])
            .sort_values(by=[samplename, Chromosome, Start_position])
            .reset_index(drop=True)
        )
    else:
        return issues


def filterAllelicFraction(maf, loc=["CGA_WES_AC"], sep=":", frac=0.1):
    """
    filters a MAF file based on allelic fraction (works with multiple sample file)

    Args:
    -----
      maf: dataframe of the maf file
      loc: list[str] colnames with the ref:alt
      sep: str separato between ref:alt
      frac: float min fraction

    Returns:
    -------
      dataframe of the maf file
    """
    muts = np.zeros((len(maf), 2))
    for val in loc:
        muts += np.array(
            [
                [v[0], 0] if "NA" in v else v
                for v in maf[val]
                .fillna("0" + sep + "0")
                .astype(str)
                .str.split(sep)
                .tolist()
            ]
        ).astype(int)
    muts = muts[:, 0] / (muts[:, 0] + muts[:, 1])
    return maf[muts >= frac]


def filterCoverage(maf, loc=["CGA_WES_AC"], sep=":", cov=4, altloc=0):
    """
    filters a MAF file based on read coverage (works with multiple sample file)

    Args:
    -----
      maf: dataframe of the maf file
      loc: list[str] colnames with the ref:alt
      sep: str separato between ref:alt
      cov: min coverage
      altloc: 0 to filter on alt and 1 to filter on ref

    Returns:
    -------
      dataframe of the maf file
    """
    muts = np.zeros((len(maf), 2))
    for val in loc:
        muts += np.array(
            [
                [v[0], 0] if "NA" in v else v
                for v in maf[val]
                .fillna("0" + sep + "0")
                .astype(str)
                .str.split(sep)
                .tolist()
            ]
        ).astype(int)
    return maf[muts[:, altloc] >= cov]


def manageGapsInSegments(
    segtocp, Chromosome="Chromosome", End="End", Start="Start", cyto=None
):
    """
    extends the ends of segments in a segment file from GATK so as to remove all gaps ove the genome (works with multiple sample file)

    Args:
    ----
      segtocp: dataframe of segments from GATK CN pipeline
      Chromosome: str the value for the Chromosome columns
      End: str the value for the End columns
      Start: str the value for the Start columns
      cyto: dataframe with chrom;end; columns giving the size of each chromosome (else puts last segment to 1000000000)
    """
    prevchr = ""
    prevend = 0
    count = 0
    l = []
    segments = segtocp.copy()
    le = len(segments)
    for k, val in segments.iterrows():
        h.showcount(count, le)
        count += 1
        if val[Chromosome] != prevchr:  # we changed chromosome
            # we extend the previous segment (last of the prev chrom) to.. way enough
            if len(l) > 0:
                l[-1][2] = (
                    1000000000
                    if cyto is None
                    else cyto[cyto["chrom"] == prevchr]["end"].values[-1]
                )
            # we extend the first segment to 0
            l.append([val[Chromosome], 0, val[End]])
        else:
            if val[Start] > prevend + 1:  # we have a gap in the same chrom
                sizeofgap = val[Start] - prevend
                # we add to the previous one half of the gap
                l[-1][2] += (
                    int(sizeofgap / 2) if sizeofgap % 2 == 0 else int(sizeofgap / 2) + 1
                )
                # the rest to the other
                l.append([val[Chromosome], val[Start] - int(sizeofgap / 2), val[End]])
            elif val[Start] < prevend:  # this should never happen
                # import pdb; pdb.set_trace()
                raise ValueError("start comes after end")
            else:
                l.append([val[Chromosome], val[Start], val[End]])
        prevchr = val[Chromosome]
        prevend = val[End]
    # we extend the last one
    l[-1][2] = (
        1000000000 if cyto is None else cyto[cyto["chrom"] == prevchr]["end"].values[-1]
    )
    segments[[Chromosome, Start, End]] = l
    return segments.reset_index(drop=True)


def toGeneMatrix(
    segments,
    gene_mapping,
    style="weighted",
    missingchrom=["Y"],
    gene_names_col="gene_name",
):
    """
    makes a geneXsample matrix from segment level copy number (works with multiple sample file)

    Args:
    ----
      style: str one of "weighted","mean","closest"
      segments: dataframe of segments containing: [Chromosome, Segment_Mean, Chromosome, start, end] columns
      gene_mapping: dataframe with symbol, ensembl_id columns for each gene
      missingchrom: list[str] chromosomes not to look into

    Returns:
    -------
      pd.dataframe: the matrix
    """
    samples = list(set(segments.DepMap_ID))
    data = np.zeros((len(samples), len(gene_mapping)))
    for i, sample in enumerate(samples):
        segs = segments[segments.DepMap_ID == sample][
            ["Chromosome", "Start", "End", "Segment_Mean"]
        ].values
        hasmissing = set(missingchrom) - set(segs[:, 0])
        j = 0
        h.showcount(i, len(samples))
        for k, gene in enumerate(gene_mapping[["Chromosome", "start", "end"]].values):
            # print(i,j)
            if gene[0] in hasmissing:
                data[i, k] = np.nan
                continue
            try:
                while gene[0] != segs[j][0] or gene[1] >= segs[j][2]:
                    # print("went beyong",gene, segs[j])
                    j += 1
                # some genes are within other genes, we need to go back in the list of segment in that case
            except:
                raise ValueError("forgot to sort one of the DF?")
            while gene[1] < segs[j][1]:
                j -= 1
                # print("decrease gene",gene)
            # we are entirely within the segment
            c = 1
            if gene[2] <= segs[j][2]:
                data[i, k] = segs[j][3]
            else:
                # how much of the gene is covered by the segment
                coef = (segs[j][2] - gene[1]) / (gene[2] - gene[1])
                # print('coef',coef)
                val = segs[j][3] * coef if style == "weighted" else segs[j][3]
                end = segs[j][2]
                # until the end of a segments goes beyond the end of the gene (say if we have X segments within the gene)
                while end < gene[2]:
                    # pdb.set_trace()
                    j += 1
                    c += 1
                    nextend = segs[j][2] if segs[j][2] < gene[2] else gene[2]
                    # here, end (of prevsegment) is the next segment's start
                    ncoef = (nextend - end) / (gene[2] - gene[1])
                    # print('multi',gene, ncoef)
                    if style == "closest":
                        if ncoef > coef:
                            val = segs[j][3]
                        else:
                            # we switch it back (see line 894)
                            ncoef = coef
                    else:
                        val += segs[j][3] * ncoef if style == "weighted" else segs[j][3]
                    end = segs[j][2]
                    coef = ncoef
                data[i, k] = val if style == "weighted" else val / c
    return pd.DataFrame(data=data, index=samples, columns=gene_mapping[gene_names_col])


def checkAmountOfSegments(segmentcn, thresh=850, samplecol="DepMap_ID"):
    """
    if there is too many segments, something might be wrong (works with multiple sample file)

    will compute the number of segments for each samples from a df of segments from RSEM

    Args:
    ----
      segmentcn: segment dataframe
      thresh: max ok amount
    """
    failed = []
    celllines = set(segmentcn[samplecol].tolist())
    amounts = []
    for cellline in celllines:
        val = segmentcn[segmentcn[samplecol] == cellline].shape[0]
        amounts.append(val)
        if val > thresh:
            failed.append(cellline)
            print(cellline, val)
    sns.kdeplot(amounts)
    return failed


def checkGeneChangeAccrossAll(genecn, thresh=0.2):
    """
    used to find poor quality genes in CN data (works with multiple sample file)

    compute given a df of gene x sample CN counts, how much change there is accross samples for
    a same gene and returns ones that are below the threshold

    Args:
    -----
      genecn: gene cn data frame
      thresh: threshold in logfold change accross all of them
    """
    return genecn.columns[genecn.var() < thresh].tolist()


def renameColumns(df):
    """
    rename some of the main columns names from RSEM, GATK.. to more readable column names
    Args:
    -----
      df: the df to rename
    Returns:
    ------
      df the renamed df
    """
    return df.rename(
        columns={
            "Sample": "DepMap_ID",
            "CONTIG": "Chromosome",
            "START": "Start",
            "END": "End",
            "seqnames": "Chromosome",
            "start": "Start",
            "end": "End",
        }
    )



# Jeremie Kalfon
# jkobject@gmail.com
# for BroadInstitute CDS
# in 2019

import time
import pandas as pd
import dalmatian as dm
import numpy as np
import os
import re
import signal
from genepy.utils import helper as h
from genepy.google import gcp
from gsheets import Sheets
from dalmatian.core import MethodNotFound

from agutil.parallel import parallelize
from agutil import status_bar
import firecloud.api
import requests
from dalmatian.core import set_timeout, DEFAULT_SHORT_TIMEOUT, APIException
import sys


def createManySubmissions(
    workspace, workflow, references, entity=None, expression=None, use_callcache=True
):
    """
    wrapper to create many submissions for a workflow

    Args:
    ----
      workspace: str namespace/workspace from url typically
      references: list(str) a list of name of the row in this entity
      entity: str terra csv type (sample_id...)
      expresson: str to use if want to compute on the direct value of the entity or on values of values
                  e.g. this.samples
      use_callcache: Bool to false if want to recompute everything even if same input

    Returns:
    ------
      submission_ids list(str) the submission ids
    """
    wm = dm.WorkspaceManager(workspace)
    submission_ids = []
    for ref in references:
        submission_ids += [
            wm.create_submission(workflow, ref, entity, expression, use_callcache)
        ]
    return submission_ids


async def waitForSubmission(workspace, submissions, raise_errors=True):
    """
    wrapper to create many submissions for a workflow

    Args:
    -----
      workspace: str namespace/workspace from url typically
      submissions: list[str] of submission ids
      raise_errors: bool to true if errors should stop your code

    Returns:
    -------
      list of ids of failed submissions
    """
    failed_submission = []
    timing = 0
    wm = dm.WorkspaceManager(workspace).disable_hound()
    assert submissions is not None
    if type(submissions) is type(""):
        submissions = [submissions]
    for scount, submission_id in enumerate(submissions):
        finished = False
        while not finished:
            done = 0
            failed = 0
            finished = True
            submission = wm.get_submission(submission_id)["workflows"]
            for _, i in enumerate(submission):
                if i["status"] not in {"Done", "Aborted", "Failed", "Succeeded"}:
                    finished = False
                elif i["status"] in {"Failed", "Aborted"}:
                    failed += 1
                    if i["workflowEntity"]["entityName"] not in failed_submission:
                        print(i["workflowEntity"]["entityName"])
                        failed_submission.append(i["workflowEntity"]["entityName"])
                elif i["status"] in {"Done", "Succeeded"}:
                    done += 1
            if not finished:
                time.sleep(40)
                print(
                    "status is: Done for "
                    + str(done)
                    + " jobs in submission "
                    + str(scount)
                    + ". "
                    + str(timing)
                    + ",5 mn elapsed.",
                    end="\r",
                )
                timing += 1
                time.sleep(20)
                print(
                    "status is: Failed for "
                    + str(failed)
                    + " jobs in submission "
                    + str(scount)
                    + ". "
                    + str(timing)
                    + " mn elapsed.",
                    end="\r",
                )
            else:
                print(
                    str(done / (done + failed))
                    + " of jobs Succeeded in submission "
                    + str(scount)
                    + "."
                )
    if len(failed_submission) > 0 and raise_errors:
        raise RuntimeError(str(len(failed_submission)) + " failed submission")
    return failed_submission
    # print and return well formated data


def removeSamples(workspace, samples):
    """
    removes a set of samples from a workspace (very usefull when we have linked pairs and pairsets)

    Args:
    -----
      workspace: str workspace name
      samples: list of samples
    """
    wm = dm.WorkspaceManager(workspace).disable_hound()
    try:
        wm.delete_sample(samples)
    except:
        print("we had pairs.")
        pairs = wm.get_pairs()
        if len(pairs) > 0:
            pairid = pairs[pairs.case_sample.isin(samples)].index.tolist()
            for k, val in wm.get_pair_sets().iterrows():
                wm.update_pair_set(k, set(val.tolist()[0]) - set(pairid))
            wm.delete_pair(pairid)
        wm.delete_sample(samples)


def uploadFromFolder(
    gcpfolder,
    prefix,
    workspace,
    sep="_",
    loc=0,
    fformat="fastq12",
    newsamples=None,
    samplesetname=None,
    source="U",
    bamcol="bam",
    baicol="bai",
    test=True,
):
    """
    upload samples (virtually: only creates tsv file) from a google bucket to a terra workspace

    A very practical function when you have uploaded a folder of samples (RNA/WES/...) in google storage
    with some naming convention, and want to have them all well listed in Terra for futher processing
    it can create a sample set.
    for a set of files: gs://bucket/path/to/files


    Args:
    -----
      gcpfolder: a gs folder path
      prefix: str the folder path
      workspace: str namespace/workspace from url typically
      sep: str the separator (only takes the first part of the name before the sep character)
      fformat bambai, fastq12, fastqR1R2 given the set of files in the folder (they need to be in this naming format)
              e.g. name.bam name.bai / name1.fastq name2.fastq / name_R1.fastq name_R2.fastq
      newsamples: DONT USE
      samplesetname: str all uploaded samples should be part of a sampleset with name..

    Returns:
    --------
      the uploaded dataframe
    """
    wm = dm.WorkspaceManager(workspace)
    print(
        "please be sure you gave access to your terra email account access to this bucket"
    )
    if samplesetname is None:
        samplesetname = "from:" + gcpfolder + prefix
    files = gcp.list_blobs_with_prefix(gcpfolder, prefix, "/")
    if fformat == "bambai":
        if newsamples is None:
            data = {"sample_id": [], bamcol: [], baicol: []}
            for file in files:
                if file.split(".")[-1] in ["bam", "bai"]:
                    name = file.split("/")[-1].split(".")[0].split(sep)[loc]
                    if name in data["sample_id"]:
                        pos = data["sample_id"].index(name)
                        if file[-4:] == ".bam":
                            data[bamcol].insert(pos, "gs://" + gcpfolder + "/" + file)
                        elif file[-4:] == ".bai":
                            data[baicol].insert(pos, "gs://" + gcpfolder + "/" + file)
                    else:
                        data["sample_id"].append(name)
                        if file[-4:] == ".bam":
                            data[bamcol].append("gs://" + gcpfolder + "/" + file)
                        elif file[-4:] == ".bai":
                            data[baicol].append("gs://" + gcpfolder + "/" + file)
                        else:
                            raise Exception("No fastq R1/R2 error", file)
                else:
                    print("unrecognized file type : " + file)
            df = pd.DataFrame(data)
            df = df.set_index("sample_id")
            print(df)
            df["participant"] = pd.Series(data["sample_id"], index=data["sample_id"])
            wm.upload_samples(df)
            wm.update_sample_set(samplesetname, df.index.values.tolist())
        else:
            # TODO: check if each column exists and can be added, else don't add it
            for i, val in enumerate(newsample["file_path"]):
                if (
                    val.split("/")[-1].split(".")[1] != "WholeGenome"
                    or val.split("/")[-2] != "bam"
                ):
                    newsample = newsample.drop(i)
                elif val.split("/")[1] != "gs:":
                    newsample["file_path"][i] = (
                        gcpfolder + newsample["file_path"][i].split("/")[-1]
                    )
            newsample = newsample.reset_index(drop=True)
            newsample = newsample.rename(
                index=str,
                columns={
                    "sample_name": "sample_id",
                    "subject_name": "participant_id",
                    "file_path": "WGS_bam",
                },
            )
            currfile = ""
            bai = [""] * int(newsample.shape[0])
            # creating an array of bai and adding it to their coresponding bams
            for i in newsample.index:
                currfile = newsample["WGS_bam"][i]
                if currfile.split("/")[-1].split(".")[-1] == "bai":
                    bai[
                        int(
                            newsample[
                                newsample["WGS_bam"] == currfile[:-4]
                            ].index.values[0]
                        )
                    ] = currfile
            newsample["WGS_bam_index"] = pd.Series(bai, index=newsample.index)
            # removing original bai rows
            for i in newsample.index:
                currfile = newsample["WGS_bam"][i]
                if currfile.split("/")[-1].split(".")[-1] == "bai":
                    newsample = newsample.drop(i)
            newsample = newsample.reset_index(drop=True)
            newsample["sample_set"] = pd.Series(
                [samplesetname] * int(newsample.shape[0]), index=newsample.index
            )
            newsample.set_index("sample_id", inplace=True, drop=True)
            newsample = newsample[
                newsample.columns.tolist()[1:] + [newsample.columns.tolist()[0]]
            ]
            newsample = newsample.loc[~newsample.index.duplicated(keep="first")]
            newsample.to_csv("temp/samples.bambai.tsv", sep="\t")
            wm.upload_samples(newsample)
            wm.update_sample_set(samplesetname, newsample.index)
    if fformat in {"fastq12", "fastqR1R2"}:
        data = {"sample_id": [], "fastq1": [], "fastq2": []}
        # print and return well formated data
        print(files)
        for file in files:
            if file[-9:] == ".fastq.gz" or file[-6:] == ".fq.gz":
                name = re.split(sep, file.split("/")[-1].split(".")[0])[loc]
                if name in data["sample_id"]:
                    pos = data["sample_id"].index(name)
                    if fformat == "fastqR1R2":
                        if "R1" in file:
                            data["fastq1"].insert(pos, "gs://" + gcpfolder + "/" + file)
                        elif "R2" in file:
                            data["fastq2"].insert(pos, "gs://" + gcpfolder + "/" + file)
                        else:
                            raise Exception("No fastq R1/R2 error", file)
                    else:
                        if file.split(".")[-3][-1] == "1":
                            data["fastq1"].insert(pos, "gs://" + gcpfolder + "/" + file)
                        elif file.split(".")[-3][-1] == "2":
                            data["fastq2"].insert(pos, "gs://" + gcpfolder + "/" + file)
                        else:
                            raise Exception("No fastq 1/2 error", file)
                else:
                    data["sample_id"].append(name)
                    if fformat == "fastqR1R2":
                        if "R1" in file:
                            data["fastq1"].append("gs://" + gcpfolder + "/" + file)
                        elif "R2" in file:
                            data["fastq2"].append("gs://" + gcpfolder + "/" + file)
                        else:
                            raise Exception("No fastq R1/R2 error", file)
                    else:
                        if file.split(".")[-3][-1] == "1":
                            data["fastq1"].append("gs://" + gcpfolder + "/" + file)
                        elif file.split(".")[-3][-1] == "2":
                            data["fastq2"].append("gs://" + gcpfolder + "/" + file)
                        else:
                            raise Exception("No fastq R1/R2 error", file)
            else:
                print("unrecognized file type : " + file)
        df = pd.DataFrame(data)
        df["Source"] = source
        df["participant"] = data["sample_id"]
        df = df.set_index("sample_id")
        if not test:
            wm.upload_samples(df)
            wm.update_sample_set(samplesetname, df.index.values.tolist())
        return df


def updateAllSampleSet(workspace, newsample_setname, Allsample_setname="all"):
    """
    update the previous All Sample sample_set with the new samples that have been added.

    It is especially useful for the aggregate task. Can more generally merge two samplesets together

    Args:
    ----
      workspace: str namespace/workspace from url typically
      newsample_setname: str name of sampleset to add to All_samples
    """
    prevsamples = list(
        dm.WorkspaceManager(workspace)
        .get_sample_sets()
        .loc[Allsample_setname]["samples"]
    )
    newsamples = list(
        dm.WorkspaceManager(workspace)
        .get_sample_sets()
        .loc[newsample_setname]["samples"]
    )
    prevsamples.extend(newsamples)
    dm.WorkspaceManager(workspace).update_sample_set(
        Allsample_setname, list(set(prevsamples))
    )


def addToSampleSet(workspace, samplesetid, samples):
    """
    add samples to a sample set

    will create new if doesn't already exist, else adds to existing

    Args:
    ----
      workspace: the workspace name
      samplesetid: the sample set name
      samples: a list of samples
    """
    try:
        prevsamples = dm.WorkspaceManager(workspace).get_sample_sets()["samples"][
            samplesetid
        ]
        samples.extend(prevsamples)
    except KeyError:
        print(
            "The sample set "
            + str(samplesetid)
            + " did not exist in the workspace. Will be created now..."
        )
    dm.WorkspaceManager(workspace).update_sample_set(samplesetid, list(set(samples)))


def addToPairSet(workspace, pairsetid, pairs):
    """
    add pairs to a pair set

    will create new if doesn't already exist, else adds to existing

    Args:
    ----
      workspace: the workspace name
      pairsetid: the pair set name
      pairs: a list of pairs
    """

    try:
        prevpairs = (
            dm.WorkspaceManager(workspace).get_pair_sets().loc[[pairsetid]].pairs[0]
        )
        pairs.extend(prevpairs)
    except KeyError:
        print(
            "The pair set "
            + str(pairsetid)
            + " did not exist in the workspace. Will be created now..."
        )
    dm.WorkspaceManager(workspace).update_pair_set(pairsetid, list(set(pairs)))


# Gwen's old version - caught some niche conditions made by get_pair_sets()
# that I think may raise errors in the current version.
# yep. the niche condition has to do with the fact that a list isn't hashable
# def addToPairSet(wm, pairsetid, pairs):
#   pairsets = wm.get_pair_sets()
#   prevpairs = pairsets.loc[[pairsetid]].pairs.tolist() # is this always a list of list? I think so.
#   print(type(prevpairs[0]))
#   if isinstance(prevpairs[0], str) :
#     pairs.extend(prevpairs)
#   elif isinstance(prevpairs[0], list):
#     pairs.extend(prevpairs[0])
#   wm.update_pair_set(pairsetid, list(set(pairs)))


def saveOmicsOutput(
    workspace,
    pathto_cnvpng="segmented_copy_ratio_img",
    pathto_stats="sample_statistics",
    specific_cohorts=[],
    specific_celllines=[],
    is_from_pairs=True,
    pathto_snv="filtered_variants",
    pathto_seg="cnv_calls",
    datadir="gs://cclf_results/targeted/kim_sept/",
    specific_samples=[],
):
    """
    *WIP* For a workspace containing all omics workflows (CNV/SNV) (like CCLF's) copies all interesting output to a data bucket

    Args:
    -----
      workspace: the workspace name
      pathto_cnvpng: sample col of the CNV plot results
      pathto_stats: sample col of the bam QC results
      specific_cohorts: if provided, will only look for this specific
      specific_celllines: if need to rrun on specific cell lines
      is_from_pairs: if we process on pairs or samples data
      pathto_snv: sample col of the snv files
      pathto_seg: sample col of the segment files
      datadir: gs bucket path where to copy the resulting files
      specific_samples: if provided will only look for these samples

    """
    if specific_cohorts:
        samples = dm.WorkspaceManager(workspace).get_samples()
        samples = samples[samples.index.isin(specificlist)]
    if is_from_pairs:
        pairs = dm.WorkspaceManager(workspace).get_pairs()
        pairs = pairs[pairs["case_sample"].isin(specificlist)]
    for i, val in samples.iterrows():
        os.system("gsutil cp " + val[pathto_seg] + " " + datadir + i + "/")
        os.system("gsutil cp " + val[pathto_cnvpng] + " " + datadir + i + "/")
        os.system("gsutil cp " + val[pathto_stats] + " " + datadir + i + "/")
        if is_from_pairs:
            snvs = pairs[pairs["case_sample"] == i][pathto_snv]
            for snv in snvs:
                if snv is not np.nan:
                    os.system("gsutil cp " + snv + " " + datadir + i + "/")
                    break
        else:
            os.system("gsutil cp " + val[pathto_snv] + " " + datadir + i + "/")


def changeGSlocation(
    workspacefrom,
    newgs,
    workspaceto=None,
    prevgslist=[],
    index_func=None,
    flag_non_matching=False,
    onlysamples=[],
    onlycol=[],
    entity="samples",
    droplists=True,
    keeppath=True,
    dry_run=True,
    par=20,
):
    """
  Function to move data around from one workspace to a bucket or to another workspace.

  can also work on dataframes containing lists of paths

  Args:
  -----
    workspacefrom: the workspace name where the data is
    newgs: the newgs bucket where to copy the data in
    workspaceto: if we should have these new samples and columns added to another workspace instead \
    of just updating the same one (usefull to copy one workspace to another)
    prevgslist: if providded, will only move files that are in the set of google bucket listed here
    index_func: *WIP* unused
    flag_non_matching: if set to true and prevgslist is set to some value, will return a list of samples that were not
    matched to anything in the prevgslist
    onlycol: do this only on a subset of columns in terra workspace
    entity: the entity in the terra workspace on which to do this
    droplists: if set to true remove all columns containing list of paths (list of path are not uploaded well in terra)
    keeppath: if set to true, will keep the full object path and just change the bucket
    dry_run: if set to true will not update anything on Terra but just return the result
    par: on how many processor do the gs copy commands.

  Returns:
  -------
    torename: the pandas.df containing the new paths
    flaglist: the samples that were non matching (if flag_non_matching is set to true)
  """
    flaglist = []
    wmfrom = dm.WorkspaceManager(workspacefrom)
    a = wmfrom.get_entities(entity)
    if len(onlysamples) > 0:
        a = a[a.index.isin(onlysamples)]
    print("using the data from " + workspacefrom + " " + entity + " list")
    if len(a) == 0:
        raise ValueError("no " + entity)
    if onlycol:
        a = a[onlycol]
    todrop = set()
    torename = {}
    print(
        'this should only contains gs:// paths otherwise precise columns using "onlycol"'
    )
    for col in a.columns.tolist():
        val = []
        for k, prev in a[col].iteritems():
            if type(prev) is str:
                new = prev
                if newgs not in new:
                    if len(prevgslist) > 0:
                        for prevgs in prevgslist:
                            new = new.replace(prevgs, newgs)
                        if flag_non_matching:
                            if new == prev:
                                flaglist.append(prev)
                    if not keeppath:
                        new = newgs + new.split("/")[-1]
                    else:
                        new = newgs + "/".join(new.split("/")[3:])
                else:
                    print("sample " + str(k) + " was already in the new gs")
                val.append(new)
            # IN CASE WE HAVE A LIST
            if type(prev) is list:
                if droplists:
                    todrop.add(k)
                    continue
                ind = []
                for prevname in prev:
                    newname = prevname
                    if newgs not in newname:
                        if len(prevgslist) > 0:
                            for prevgs in prevgslist:
                                new = new.replace(prevgs, newgs)
                            if flag_non_matching:
                                if new == prev:
                                    flaglist.append(prev)
                        if not keeppath:
                            new = newgs + new.split("/")[-1]
                        else:
                            new = newgs + "/".join(new.split("/")[3:])
                    else:
                        print("sample " + str(k) + " was already in the new gs")
                    ind.append(newname)
                val.append(ind)
        torename.update({col: val})
        if not dry_run:
            if keeppath:
                h.parrun(
                    [
                        "gsutil mv " + a.iloc[i][col] + " " + v
                        for i, v in enumerate(val)
                    ],
                    cores=20,
                )
            else:
                gcp.mvFiles(a[col].tolist(), newgs)
        else:
            if keeppath:
                print(
                    ["gsutil mv " + a.iloc[i][col] + " " + v for i, v in enumerate(val)]
                )
            else:
                print("mv " + str(a[col].tolist()) + " " + newgs)
    torename = pd.DataFrame(
        data=torename, index=[i for i in a.index.tolist() if i != "nan"]
    )
    if workspaceto is not None:
        wmto = dm.WorkspaceManager(workspaceto)
        if not dry_run:
            wmto.disable_hound().update_entity_attributes(entity, torename)
    return torename, flaglist


def renametsvs(workspace, wmto=None, index_func=None):
    """
    ################## WIP ############
    only works for one use case
    """
    data = {}
    wmfrom = dm.WorkspaceManager(workspace)
    try:
        a = wmfrom.get_participants()
        data.update({"participants": a})
    except:
        print("no participants")
    try:
        a = wmfrom.get_samples()
        data.update({"samples": a})
    except:
        print("no samples")
    try:
        a = wmfrom.get_pair_sets()
        data.update({"pair_sets": a})
    except:
        print("no pair_sets")
    try:
        a = wmfrom.get_pairs()
        data.update({"pairs": a})
    except:
        print("no pairs")
    try:
        a = wmfrom.get_sample_sets()
        data.update({"sample_sets": a})
    except:
        print("no sample_sets")
    # currently works only for sample, sample
    for k, entity in data.items():
        ind = []
        for i in entity.index:
            pos = val.find("-SM")
            if pos != -1:
                val = val[pos + 1 :]
                pos = val.find("-SM")
                if pos != -1:
                    val = val[:9] + val[pos + 1 :]
            ind.append(val)
        entity.index = ind
        # for all columns of the tsv
        for k, val in entity.iterrows():
            for i, v in enumerate(val):
                if type(v) is list or type(v) is str:
                    ind = []
                    for j in v:
                        pos = j.find("-SM")
                        if pos != -1:
                            j = j[pos + 1 :]
                            pos = j.find("-SM")
                            if pos != -1:
                                j = j[:9] + j[pos + 1 :]
                        ind.append(j)
                    val[i] = ind
                entity.loc[k] = val
        if wmto is None:
            wmto = wmfrom
        if "participants" in data:
            wmto.upload_participants(data["participants"].index.tolist())
        if "samples" in data:
            wmto.upload_samples(data["samples"])
        if "pairs" in data:
            wmto.upload_entities("pair", data["pairs"])
        if "pair_set" in data:
            pairset = data["pair_set"].drop("pairs", 1)
            wmto.upload_entities("pair_set", pairset)
            for i, val in data["pair_set"].iterrows():
                wmto.update_pair_set(i, val.pairs)
        if "sample_set" in data:
            sampleset = data["sample_set"].drop("samples", 1)
            wmto.upload_entities("sample_set", sampleset)
            for i, val in data["sample_set"].iterrows():
                wmto.update_sample_set(i, val.samples)


async def findBackErasedDuplicaBamteFromTerraBucket(
    workspace, gsfolder, bamcol="WES_bam", baicol="WES_bai"
):
    """
    If you have erased bam files in gcp with bai files still present and the bam files are stored elsewhere
    and their location is in a terra workspace.

    Will find them back by matching bai sizes and copy them back to their original locations

    Args:
    ----
      workspace: str namespace/workspace from url typically
      gsfolder: str the gsfolder where the bam files are
      bamcol: str colname of the bam
      baicol: str colname of the bai
    """
    # get ls of all files folder
    samples = os.popen("gsutil -m ls -al " + gsfolder + "**.bai").read().split("\n")
    # compute size filepath

    sizes = {
        "gs://" + val.split("gs://")[1].split("#")[0]: int(val.split("2019-")[0])
        for val in samples[:-2]
    }
    names = {}
    for k, val in sizes.items():
        if val in names:
            names[val].append(k)
        else:
            names[val] = [k]
    # get all bai in tsv
    samp = dm.WorkspaceManager(workspace).get_samples()
    for k, val in samp.iterrows():
        if val[bamcol] != "NA" and val[baicol] != "NA":
            # if bai has duplicate size
            code = os.system("gsutil ls " + val[bamcol])
            if code == 256:
                if val[bamcol] is None:
                    print("we dont have bam value for " + str(k))
                    continue
                else:
                    print("no match values for " + str(val[bamcol]))

                for va in names[sizes[val[baicol]]]:
                    # for all duplicate size
                    # if ls bam of bai duplicate size work
                    # mv bam to bampath in gsfolder
                    if ".bam" in va:
                        if (
                            os.system("gsutil ls " + va.split(".bam.bai")[0] + ".bam")
                            == 0
                        ):
                            print(
                                "gsutil mv "
                                + va.split(".bam.bai")[0]
                                + ".bam "
                                + val[bamcol]
                            )
                            os.system(
                                "gsutil mv "
                                + va.split(".bam.bai")[0]
                                + ".bam "
                                + val[bamcol]
                            )
                            break
                    elif os.system("gsutil ls " + va.split(".bai")[0] + ".bam") == 0:
                        print(
                            "gsutil mv " + va.split(".bai")[0] + ".bam " + val[bamcol]
                        )
                        os.system(
                            "gsutil mv " + va.split(".bai")[0] + ".bam " + val[bamcol]
                        )
                        break
            elif code == signal.SIGINT:
                print("Awakened")
                break
        else:
            print("no data for " + str(k))


async def shareTerraBams(
    samples,
    users,
    workspace,
    bamcols=["internal_bam_filepath", "internal_bai_filepath"],
    unshare=False,
):
    """
    will share some files from gcp with a set of users using terra as metadata repo.

    only works with files that are listed on a terra workspace tsv but actually
    point to a regular google bucket and not a terra bucket.

    Args:
    ----
      users: list[str] of users' google accounts
      workspace: str namespace/workspace from url typically
      samples list[str] of samples_id for which you want to share data
      bamcols: list[str] list of column names of gsfiles to share

    Returns:
    --------
      a list of the gs path we have been giving access to
    """
    if type(users) is str:
        users = [users]
    wm = dm.WorkspaceManager(workspace)
    togiveaccess = np.ravel(wm.get_samples()[bamcols].loc[samples].values)
    key = "-rd " if unshare else "-ru "
    for user in users:
        files = ""
        for i in togiveaccess:
            files += " " + i
        code = os.system(
            "gsutil -m acl ch " + key + user + (" " if unshare else ":R ") + files
        )
        if code == signal.SIGINT:
            print("Awakened")
            break
    print("the files are stored here:\n\n")
    print(togiveaccess)
    print("\n\njust install and use gsutil to copy them")
    print("https://cloud.google.com/storage/docs/gsutil_install")
    print("https://cloud.google.com/storage/docs/gsutil/commands/cp")
    return togiveaccess


def saveWorkspace(workspace, folderpath):
    """
    will save everything about a workspace into a csv and json file

    Args:
    -----
      workspace: str namespace/workspace from url typically
        namespace (str): project to which workspace belongs
        workspace (str): Workspace name
      folderpath: str path to save files
    """
    wm = dm.WorkspaceManager(workspace)
    h.createFoldersFor(folderpath)

    conf = wm.get_configs()
    for k, val in conf.iterrows():
        with open(folderpath + val["name"] + ".wdl", "w") as f:
            if val.sourceRepo == "dockstore":
                name = (
                    "dockstore.org/"
                    + "/".join(val["methodPath"].split("/")[2:4])
                    + "/"
                    + val["methodVersion"]
                )
            else:
                name = "/".join(
                    val[["methodNamespace", "methodName", "methodVersion"]]
                    .astype(str)
                    .tolist()
                )
            try:
                f.write(dm.get_wdl(name))
            except MethodNotFound:
                print(name + " could not be found")
    conf.to_csv(folderpath + "worflow_list.csv")
    params = {}
    params["GENERAL"] = wm.get_workspace_metadata()
    for k, val in conf.iterrows():
        params[k] = wm.get_config(val["name"])
        h.dictToFile(
            params[k]["inputs"], folderpath + "inputs_" + val["name"] + ".json"
        )
        h.dictToFile(params[k], folderpath + "conf_" + val["name"] + ".json")
        h.dictToFile(
            params[k]["outputs"], folderpath + "outputs_" + val["name"] + ".json"
        )
    h.dictToFile(params, folderpath + "all_configs.json")


async def cleanWorkspace(
    workspaceid,
    only=[],
    toleave=[],
    defaulttoleave=[
        "workspace",
        "scripts",
        "notebooks",
        "files",
        "data",
        "hound",
        "references",
        "name",
        "folder",
    ],
):
    """
    removes all processing folder in a terra workspace easily

    args:
      only: list of strings to keep
      workspaceid: str, the workspace
      toleave: a list of first order folder in the bucket that you don't want to be deleted
      defaulttoleave: it should contain non processing folders that contain metadata and files for the workspace
    """
    toleave.extend(defaulttoleave)
    bucket = dm.WorkspaceManager(workspaceid).get_bucket_id()
    res = subprocess.run("gsutil -m ls gs://" + bucket, shell=True, capture_output=True)
    if res.returncode != 0:
        raise ValueError(str(res.stderr))
    res = str(res.stdout)[2:-1].split("\\n")[:-1]
    toremove = [val for val in res if val.split("/")[-2] not in toleave]
    if only:  # you were here
        toremove = [val for val in res if val.split("/")[-2] in only]
    if h.askif(
        "we are going to remove "
        + str(len(toremove))
        + " files/folders:\n"
        + str(toremove)
        + "\nare you sure?"
    ):
        gcp.rmFiles(toremove, add="-r")
    else:
        print("aborting")


def changeToBucket(
    samples,
    gsfolderto,
    name_col=None,
    values=["bam", "bai"],
    filetypes=None,
    catchdup=False,
    dryrun=True,
):
    """
    moves all bam/bai files in a sampleList from Terra, to another gs bucket and rename them in the sample list

    will prevent erasing a duplicate sample by adding a random string or by flagging them and not copying them

    Args:
    ----
      samples: pandas.dataframe with columns to move
      gsfolderto: the bucket path to move the data to
      values: list of the cols in the dataframe containing the gs object path to be moved
      filetypes: list[str] of size values for each columns, give a suffix (.txt, .bam, ...)
      catchdup: if false will prepend a random string to the names before moving them, else will flag duplicate names
      dryrun: only shows the output but does not move the files

    Returns:
    --------
      the updated sample pandas.dataframe
    """
    # to do the download to the new dataspace
    for i, val in samples.iterrows():
        ran = h.randomString(6, "underscore", withdigits=False)
        for j, ntype in enumerate(values):
            # TODO try:catch
            filetype = (
                ".".join(val[ntype].split("/")[-1].split(".")[1:])
                if filetypes is None
                else filetypes[j]
            )
            if name_col is None:
                name = val[ntype].split("/")[-1].split(".")[0]
            elif name_col == "index":
                name = val.name
            else:
                name = val[name_col]
            name = (
                name + "." + filetype if catchdup else name + "_" + ran + "." + filetype
            )
            if not gcp.exists(gsfolderto + name) or not catchdup:
                cmd = "gsutil cp " + val[ntype] + " " + gsfolderto + name
                if dryrun:
                    print(cmd)
                else:
                    res = subprocess.run(cmd, shell=True, capture_output=True)
                    if res.returncode != 0:
                        raise ValueError(str(res.stderr))
                    samples.loc[i, ntype] = gsfolderto + name
            else:
                print(name + " already exists in the folder: " + gsfolderto)
                print(gcp.lsFiles([gsfolderto + name], "-la"))
    return samples


def deleteJob(workspaceid, subid, taskid, deleteCurrent=False, dryrun=True):
    """
    removes files generated by a job on Terra

    Args:
    -----
      workspaceid: str wokspace name
      subid: str the name of the job
      taskid: str the name of the task in this job
      DeleteCurrent: bool whether or not to delete files if they appear in one of the sample/samplesets/pairs data tables
      dryrun: bool just plot the commands but don't execute them
    """
    wm = dm.WorkspaceManager(workspaceid)
    bucket = wm.get_bucket_id()
    data = []
    if deleteCurrent:
        if dryrun:
            print("gsutil -m rm gs://" + bucket + "/" + subid + "/*/" + taskid + "/**")
        else:
            res = subprocess.run(
                "gsutil -m rm gs://" + bucket + "/" + subid + "/*/" + taskid + "/**",
                shell=True,
                capture_output=True,
            )
            if res.returncode != 0:
                raise ValueError(str(res.stderr))
    else:
        res = subprocess.run(
            "gsutil -m ls gs://" + bucket + "/" + subid + "/*/" + taskid + "/**",
            shell=True,
            capture_output=True,
        )
        if res.returncode != 0 or len(str(res.stdout)) < 4:
            raise ValueError(str(res.stderr))
        data += str(res.stdout)[2:-1].split("\\n")[:-1]
        if "TOTAL:" in data[-1]:
            data = data[:-1]
        sam = pd.concat([wm.get_samples(), wm.get_pairs(), wm.get_sample_sets()])
        tokeep = set(
            [
                val
                for val in sam.values.ravel()
                if type(val) is str and val[:5] == "gs://"
            ]
        )
        torm = set(data) - tokeep
        if dryrun:
            print(torm)
        else:
            h.parrun(["gsutil rm " + i for i in torm], cores=12)


# removing things from old failed workflows
def removeFromFailedWorkflows(
    workspaceid, maxtime="2020-06-10", everythingFor=[], dryrun=False
):
    """
    Lists all files from all jobs that have failed and deletes them.

    Can be very long

    Args:
    -----
      workspaceid: str the workspace name
      maxtime: str date format (eg. 2020-06-10) does not delete files generated past this date
      everythingFor: list[str] removes from these workflows even if not failed
      dryrun: bool whether or not to execute or just print commands
    """
    wm = dm.WorkspaceManager(workspaceid)
    for k, val in wm.get_submission_status(filter_active=False).iterrows():
        if (
            val.Failed > 0 or val.configuration in everythingFor
        ) and val.date.date() > pd.to_datetime(maxtime):
            for w in wm.get_submission(val.submission_id)["workflows"]:
                if w["status"] == "Failed" or val.configuration in everythingFor:
                    try:
                        a = w["workflowId"]
                    # else it was not even run
                    except:
                        continue
                    deleteJob(workspaceid, val.submission_id, a, dryrun=dryrun)


async def deleteHeavyFiles(workspaceid, unusedOnly=True):
    """
    deletes all files above a certain size in a workspace (that are used or unused)

    Args:
    ----
      workspaceid: str the name off the workspace
      unusedOnly: bool whether to delete used files as well (files that appear in one of the sample/samplesets/pairs data tables)
    """
    wm = dm.WorkspaceManager(workspaceid)
    bucket = wm.get_bucket_id()
    sizes = gcp.get_all_sizes("gs://" + bucket + "/")
    print("we got " + str(len(sizes)) + " files")
    a = list(sizes.keys())
    a.sort()
    ma = 100
    torm = []
    tot = 0
    for i in a[::-1]:
        if i > 1000000 * ma:
            tot += i
            for val in sizes[i]:
                torm.append(val)
    print("we might remove more than " + str(tot / 1000000000) + "GB")
    if unusedOnly:
        sam = pd.concat([wm.get_samples(), wm.get_pairs(), wm.get_sample_sets()])
        tokeep = set(
            [
                val
                for val in sam.values.ravel()
                if type(val) is str and val[:5] == "gs://"
            ]
        )
        torm = set(torm) - tokeep
    return torm


def findFilesInWorkspaces(names=[], lookup=["**", "*.", ".*"]):
    """
    given All your terra workspaces, find a given gs filename

    Args:
    -----
      names: list[str] of filenames to find
      lookup: list[str] a set of flags giving how to look
        [** through all folders, *. can be preprended with anything,
        .* can be appended with anything]
    """
    ws = dm.list_workspaces()
    print("listing workspacs")
    file = []
    res = []
    for val in ws:
        val = val["workspace"]
        print(val["namespace"] + "/" + val["name"])
        buck = "gs://" + val["bucketName"] + "/"
        if (
            subprocess.run(
                "gsutil ls " + buck, capture_output=True, shell=True
            ).returncode
            != 0
        ):
            print("cannot access this bucket")
            continue
        if len(names) == 0:
            file.append(buck)
        if "**" in lookup:
            buck += "**"
        if not "*." in lookup:
            buck += "/"
        for name in names:
            val = buck + name
            if ".*" in lookup:
                val += "*"
            data = subprocess.run(
                "gsutil -m ls " + val, capture_output=True, shell=True
            )
            if data.returncode != 0:
                if "One or more URLs matched no objects" not in str(data.stderr):
                    raise ValueError("issue with the command: " + str(data.stderr))
            if len(str(data.stdout)) < 4:
                continue
            res += str(data.stdout)[2:-1].split("\\n")[:-1]
            if "TOTAL:" in res[-1]:
                res = res[:-1]
    return res


def uploadWorkflows(workspaceID, workflows, path=None):
    """
    updates the workflows on Terra and upgrades the workflow values on our workspace

    Args:
    -----
        workflows: dict(workflowID,location) or list(workflowID) if path
        path: folder path where files with same name as workflows' name are stored
    """
    method_folder = "src/"
    methods = [""]
    dm.update_method()


def update_entities(workspace, etype, oetype, target_set=None):
    """Attach entities (samples or pairs) to participants

    Args:
    -----
        workspace: str the workspace name
        etype: str the entity type to attach
        oetype: str the entity type to attach to
        target_set: str the set to attach to. If None, won't attach to a set
    """
    wm = dm.WorkspaceManager(workspace)
    # get etype -> participant mapping
    df = wm.get_entities(etype)
    if oetype + "_id" not in df.columns:
        raise ValueError("{}_id column not found in samples".format(oetype))
    df = df[[oetype + "_id"]].astype(str)

    entities_dict = {k: g.index.values for k, g in df.groupby(oetype + "_id")}
    oe_ids = np.unique(df[oetype + "_id"])

    column = "{}s_{}".format(etype, (target_set if target_set is not None else ""))
    last_result = 0

    @parallelize(4)
    def update_mc(oe_id):
        attr_dict = {
            column: {
                "itemsType": "EntityReference",
                "items": [
                    {"entityType": etype, "entityName": i} for i in entities_dict[oe_id]
                ],
            }
        }
        attrs = [firecloud.api._attr_set(i, j) for i, j in attr_dict.items()]
        # It adds complexity to put the context manager here, but
        # since the timeout is thread-specific it needs to be set within
        # the thread workers
        with set_timeout(DEFAULT_SHORT_TIMEOUT):
            try:
                r = firecloud.api.update_entity(
                    wm.namespace, wm.workspace, oetype, oe_id, attrs
                )
                if r.status_code != 200:
                    last_result = r.status_code
            except requests.ReadTimeout:
                return oe_id, 503  # fake a bad status code to requeue
        return oe_id, r.status_code

    n_oes = len(oe_ids)

    with wm.hound.batch():
        with wm.hound.with_reason(
            "<Automated> Populating attribute from entity references"
        ):
            for attempt in range(5):
                retries = []

                for k, status in status_bar.iter(
                    update_mc(oe_ids),
                    len(oe_ids),
                    prepend="Updating {}s for {} ".format(etype, oetype),
                ):
                    if status >= 400:

                        retries.append(k)
                    else:
                        wm.hound.update_entity_meta(
                            oetype, k, "Updated {} membership".format(column)
                        )
                        wm.hound.update_entity_attribute(
                            oetype, k, column, list(entities_dict[k])
                        )

                if len(retries):
                    if attempt >= 4:
                        print(
                            "\nThe following",
                            len(retries),
                            " " + oetype + " could not be updated:",
                            ", ".join(retries),
                            file=sys.stderr,
                        )
                        raise APIException(
                            "{} {} could not be updated after 5 attempts".format(
                                len(retries), oetype
                            ),
                            last_result,
                        )
                    else:
                        print("\nRetrying remaining", len(retries), " " + oetype)
                        oe_ids = [item for item in retries]
                else:
                    break

        print("\n    Finished attaching {}s to {} {}".format(etype, n_oes, oetype))



import requests
from collections import defaultdict, Counter
import pandas as pd
from io import StringIO

csv_url = 'https://intranet.broadinstitute.org/~datasci/cell_lines/name_mapping.csv'
__version__ = "0.1"

def read_file_to_dict(key_name, value_name):
    return_dict = defaultdict(str)

    response = requests.get(csv_url)
    assert response.status_code == 200, "Could not fetch mapping from {}, got, status_code={} reason={}".format(csv_url, response.status_code, response.reason)
    mapping_request = StringIO(response.text)
    df = pd.read_csv(mapping_request)

    for row in df.to_dict(orient='records'):
        if return_dict[row[key_name]] == "":
            return_dict[row[key_name]] = row[value_name]
        else:
            if row[value_name] != return_dict[row[key_name]]:      
                return_dict[row[key_name]] += str(", " + row[value_name])

    return return_dict


def check_unique(input_list, result_list):
    input_len = len(input_list)
    # make sure one key doesn't point to multiple values
    is_unique = True
    for mapping in result_list:
        if mapping is not None and ',' in mapping:
            is_unique = False
            break

    # make sure two keys don't point to the same value
    if is_unique:
        num_unique_values = len(Counter(tuple(mapping) for mapping in result_list if mapping is not None)) + int(None in result_list)

        is_unique = (input_len == num_unique_values)

    if not is_unique:
        raise RuntimeError('Mappings are not unique')


def name_mapper(input_type, input_names, output_type, ignore_problems=False, check_unique_mapping=True):
    if len(input_names) == 0:
        raise RuntimeError("Please input a non-empty list")

    mapping_dict = read_file_to_dict(input_type, output_type)
    output_names = []

    for i_name in input_names:
        o_name = mapping_dict.get(i_name)
        if not ignore_problems and o_name is None:
            raise KeyError(output_type + " " + "could not be found for the following " + input_type + ":  " + i_name)
        output_names.append(o_name)

    if check_unique_mapping:
        check_unique(input_names, output_names)

    assert len(output_names) == len(input_names)
    return output_names


def arxspan_to_ccle(arxspan_ids, ignore_problems=False, check_unique_mapping=True):
    return name_mapper('broad_id', arxspan_ids, 'canonical_ccle_name', ignore_problems, check_unique_mapping)


def ccle_to_arxspan(ccle_names, ignore_problems=False, check_unique_mapping=True):
    return name_mapper('ccle_name', ccle_names, 'broad_id', ignore_problems, check_unique_mapping)


def ccle_to_latest(ccle_names, ignore_problems=False, check_unique_mapping=True):
    return name_mapper('ccle_name', ccle_names, 'canonical_ccle_name', ignore_problems, check_unique_mapping)

# alias for the old name of this function
latest_ccle_names=ccle_to_latest



# 1. make sure dalmatian is installed
# pip install firecloud-dalmatian
# 2. make sure graphviz is installed (this script will run "dot" to generate images)
# brew install graphviz

import subprocess
import requests
import pandas as pd
import dalmatian
import subprocess


def resolve_dot_path(root_entity_type, name):
  name = name.replace(" ", "")
  parts = name.split(".")
  assert parts[0] == "this"
  del parts[0]
  cur = root_entity_type
  while len(parts) > 1:
    next_part = parts[0]
    if cur == "sample_set" and next_part == "samples":
      cur = "sample"
    elif cur == "pair" and next_part == "case_sample":
      cur = "sample"
    elif cur == "pair" and next_part == "control_sample":
      cur = "sample"
    elif cur == "sample" and next_part == "participant":
      cur = "participant"
    else:
      raise Exception(f"Unknown case: {cur} -> {next_part} (root_entity_type={root_entity_type} name={name})")
    del parts[0]
  return cur+"."+parts[0]

def extract_config_summary(workspace_name, workflows=None):
  wm = dalmatian.WorkspaceManager(workspace_name)
  configs = wm.get_configs()

  config_summaries = []
  for rec in configs.to_records():
    cfgname = rec['namespace']+"/"+rec['name']
    if workflows is not None:
      if cfgname not in workflows:
        continue
    config = wm.get_config(cfgname)
    config['inputs'] = {k:v.strip() for k,v in config['inputs'].items()}
    config['outputs'] = {k:v.strip() for k,v in config['outputs'].items()}
    inputs=[resolve_dot_path(config['rootEntityType'], x) for x in config['inputs'].values() if x.startswith("this.")]
    outputs=[resolve_dot_path(config['rootEntityType'], x) for x in config['outputs'].values() if x.startswith("this.")]
    config_summaries.append(dict(inputs=inputs, outputs=outputs, entity_type=rec['rootEntityType'], name=cfgname))
  return config_summaries


def write_dependency_graph_image(filename, config_summaries):
  with open("/tmp/sample.dot", "wt") as fd:
    node_names = {}
    def nn(name, is_var):
      if name in node_names:
        return node_names[name]
      node_name = "n{}".format(len(node_names))
      node_names[name] = node_name
      fd.write("{} [label=\"{}\" {}];\n".format(node_name, name, {True: "shape=oval", False: "shape=box fillcolor=yellow style=filled"}[is_var]))
      return node_names[name]

    fd.write("digraph { rankdir=LR;\n")
    for config in config_summaries:
      for name in config['inputs']:
        fd.write("{} -> {};\n".format(nn(name, True), nn(config['name'], False)))

      for name in config['outputs']:
        fd.write("{} -> {};\n".format(nn(config['name'], False), nn(name, True)))
    fd.write("}\n")


  subprocess.check_call(["dot", "/tmp/sample.dot", "-Tpng", "-o", filename])

def write_config_summary_table(filename, config_summaries):
  from collections import defaultdict
  variables = defaultdict(lambda: dict(used_by=[], produced_by=[]))
  for config in config_summaries:
    for name in config['inputs']:
      variables[name]['used_by'].append(config['name'])

    for name in config['outputs']:
      variables[name]['produced_by'].append(config['name'])

  var_col = []
  used_by = []
  produced_by = []
  inputs = []
  for k, v in variables.items():
    if len(v["produced_by"]) == 0:
      inputs.append(k)
    var_col.append(k)
    used_by.append(", ".join(v["used_by"]))
    produced_by.append(", ".join(v["produced_by"]))
  df = pd.DataFrame(dict(variable=var_col, used_by=used_by, produced_by=produced_by))
  df.to_csv(filename)

def map_workspace_diagram(workspace_name, output_path='terra-workflows', workflows=None):
  """
  -- adapted from scripts written by Philip Montgomery --
  this function creates a graph of the workflows within a Terra workspace

  inputs:
      workspace_name (str): name of the workspace
      output_path (str): path where outputs will be saved
      workflows (list[str]): list of workflows to consider in the graph. If None is provided (default),
          all the workflows in the workspace will be used.

  example code:
      workflows = ['stewart/pipette_wgs_SV', 'stewart/manta', 'stewart/SvABA_xtramem',
              'stewart/svaba_snowmanvcf2dRangerForBP', 'stewart/mantavcf2dRangerForBP',
              'stewart/extract_dRanger_intermediates','stewart/pcawg_snowmanvcf2dRangerForBP',
              'stewart/SV_cluster_forBP', 'stewart/breakpointer', 'stewart/Breakpointer_fix_sample',
              'stewart/REBC_SV_consensus_filter_v3']

      workspace_name = 'broad-firecloud-ccle/REBC_methods_only-tmp'
      print(workspace_name)
      config = mtw.map_workspace_diagram(workspace_name, workflows=workflows)
  """
  configs = extract_config_summary(workspace_name, workflows=workflows)
  write_dependency_graph_image(output_path+'/'+workspace_name.replace("/", " ")+".png", configs)
  write_config_summary_table(output_path+'/'+workspace_name.replace("/", " ")+".csv", configs)
  return configs



# Jeremie Kalfon
# for BroadInsitute
# in 2019

from __future__ import print_function
import venn as pyvenn
from matplotlib import pyplot as plt
from matplotlib import cm
from bokeh.palettes import *
from bokeh.layouts import layout, widgetbox, column, row
from bokeh.models.annotations import LabelSet
from bokeh.models.widgets import TextInput
from bokeh.models import HoverTool, CustomJS, BasicTicker, ColorBar, PrintfTickFormatter
from bokeh.models import ColumnDataSource, LinearColorMapper, LogColorMapper
from bokeh.util.hex import hexbin
from bokeh.transform import linear_cmap
from bokeh.io import show, export_svg
from bokeh.plotting import *
import bokeh
import colorcet as cc
from PIL import Image, ImageDraw, ImageFont
import seaborn as sns

import pandas as pd
from math import pi
import numpy as np


def scatter(
    data,
    labels=None,
    title="scatter plot",
    showlabels=False,
    folder="",
    colors=None,
    xname="",
    yname="",
    importance=None,
    radi=5,
    alpha=0.8,
    colprovided=False,
    shape=None,
    no_xy=False,
    **kwargs
):
    """
    Makes an interactive scatter plot using Bokeh

    Args:
    -----
        data: array-like with shape [N,2]
        labels: list[str] a list of N names for each points or dict[str, list[str]] if you want multiple labels
        title: str the plot title
        showlabels: bool if the labels should be always displayed or not (else just on hover)
        colors: list[int] of N integers from 0 up to 256 for the dot's colors
        folder: str of location where to save the plot, won't save if empty
        xname: str the name of the x axes
        yname: str the name of the y axes
        importance: a list[int] of N values to scale the size of the dots and their opacity by
        radi: int the size of the dots
        alpha: float the opacity of the dots
        **kwargs: additional bokeh.figure args

    Returns:
    ------
        the bokeh object
    """
    TOOLS = "hover,crosshair,pan,wheel_zoom,zoom_in,zoom_out,box_zoom,undo,redo,reset,save,box_select,lasso_select,"

    radii = []
    fill_alpha = []
    cols = []

    col = viridis(len(set(colors))) if colors is not None else ["#29788E"]  # (viridis1)
    for i in range(data.shape[0]):
        radii.append(radi if importance is None else radi / (1 + importance[i]))
        fill_alpha.append(
            alpha if importance is None else alpha - (0.2 * importance[i])
        )
        if not colprovided:
            cols.append(col[0] if colors is None else col[int(colors[i])])

    if shape is None:
        shape = [0] * data.shape[0]
    shape = np.array(shape)

    TOOLTIPS = (
        []
        if no_xy
        else [
            ("(x,y)", "(@x, @y)"),
        ]
    )
    if type(labels) is list:
        TOOLTIPS.append(("label", "@labels"))
    elif type(labels) is dict:
        TOOLTIPS.extend([(str(i), "@" + str(i)) for i in list(labels.keys())])

    p = figure(tools=TOOLS, tooltips=TOOLTIPS, title=title, output_backend="webgl")

    shaplot = {
        0: p.circle,
        1: p.star,
        2: p.square,
        3: p.triangle,
        6: p.diamond,
        7: p.inverted_triangle,
        5: p.hex,
        4: p.cross,
    }

    shape_encoder = {val: i for i, val in enumerate(set(shape))}
    if len(set(shape)) > 8:
        raise ValueError("Too many shapes")
    for val in set(shape):
        di = dict(
            x=data[shape == val, 0],
            y=data[shape == val, 1],
            fill_color=np.array(cols)[shape == val]
            if not colprovided
            else np.array(colors)[shape == val],
            fill_alpha=np.array(fill_alpha)[shape == val],
            radius=np.array(radii)[shape == val],
        )

        if type(labels) is list:
            di.update({"labels": np.array(labels)[shape == val]})
        elif type(labels) is dict:
            di.update(
                {va: np.array(label)[shape == val] for va, label in labels.items()}
            )

        source = ColumnDataSource(data=di)

        shaplot[shape_encoder[val]](
            "x",
            "y",
            color="fill_color",
            fill_alpha="fill_alpha",
            line_alpha="fill_alpha",
            line_width=1,
            size="radius",
            source=source,
        )
    p.xaxis[0].axis_label = xname
    p.yaxis[0].axis_label = yname
    if showlabels:
        labels = LabelSet(
            x="x",
            y="y",
            text="labels",
            level="glyph",
            text_font_size="9pt",
            x_offset=5,
            y_offset=5,
            source=source,
            render_mode="canvas",
        )
        p.add_layout(labels)
    try:
        show(p)
    except:
        show(p)
    if folder:
        save(p, folder + title.replace(" ", "_") + "_scatter.html")
        try:
            p.output_backend = "svg"
            export_svg(p, filename=folder + title.replace(" ", "_") + "_scatter.svg")
        except RuntimeError:
            print("You need to install Selenium to save the svg!")
    return p


def bigScatter(
    data,
    precomputed=False,
    logscale=False,
    features=False,
    title="BigScatter",
    binsize=0.1,
    folder="",
    showpoint=False,
):
    """
    uses a binning method to display millions of points at the same time and showcase density.

    Does this in an interactive fashion

    Args:
    -----
        data: array like. the array containing the point location x,y or their location and density of bins (q,r,c)
        precomputed: bool whether or not the array has aleady been hexbined
        logscale: bool, whether or not the data is logscaled
        features: list[] if the matrix contains a feature column with feature information (names, values. for each bin/dot)
        title: str the title of the plot
        binsize: float the size of the bins
        showpoint: bool whether or not to display them as points or hexes.
        folder: str of location where to save the plot, won't save if empty

    Returns:
    ------
        the bokeh object
    """
    TOOLS = "wheel_zoom,zoom_in,zoom_out,box_zoom,undo,redo,reset,save,box_select,lasso_select,"
    names = [("count", "@c")]
    if features:
        names.append(("features", "@features"))
    if precomputed:
        TOOLS = "hover," + TOOLS
    p = figure(
        title=title,
        tools=TOOLS,
        tooltips=names if precomputed else None,
        match_aspect=True,
        background_fill_color="#440154",
    )
    if precomputed:
        p.hex_tile(
            q="q",
            r="r",
            size=binsize,
            line_color=None,
            source=data,
            hover_color="pink",
            hover_alpha=0.8,
            fill_color=linear_cmap("c", "Viridis256", 0, max(data.c))
            if not logscale
            else {"field": "c", "transform": LogColorMapper("Viridis256")},
        )
    else:
        if features:
            print("we cannot yet process features on non precomputed version")
        r, bins = p.hexbin(
            data[:, 0],
            data[:, 1],
            line_color=None,
            size=binsize,
            hover_color="pink",
            hover_alpha=0.8,
            fill_color=linear_cmap("c", "Viridis256", 0, None)
            if not logscale
            else {"field": "c", "transform": LogColorMapper("Viridis256")},
        )
    p.grid.visible = False
    if showpoint:
        p.circle(data[:, 0], data[:, 1], color="white", size=1)

    if not precomputed:
        p.add_tools(
            HoverTool(
                tooltips=names,
                mode="mouse",
                point_policy="follow_mouse",
                renderers=[r] if not precomputed else None,
            )
        )
    p.output_backend = "svg"
    try:
        show(p)
    except:
        show(p)
    if folder:
        save(p, folder + title.replace(" ", "_") + "_scatter.html")
        try:
            export_svg(p, filename=folder + title.replace(" ", "_") + "_scatter.svg")
        except RuntimeError:
            print("You need to install Selenium to save the svg!")
    return p


def CNV_Map(
    df,
    sample_order=[],
    title="CN heatmaps sorted by SMAD4 loss, pointing VPS4B",
    width=900,
    height=400,
    standoff=10,
    y_label="",
    marks=[],
):
    """
    create an interactive plot suited for visualizing segment level CN data for a set of samples using bokeh

    args:
    ----
        df: df['Sample' 'Start' 'End' 'Segment_Mean' 'size'] the df containing segment level
        copy number (can be subsetted to a specific region or genome-wide)
        sampleorder: list[Sample] <- for all samples present in the df
        title: plot title
        width: int width
        height: int height
        standoff: the space between the plot and the x axis
        y_label: the y axis label
        marks: location of lines at specific loci

    Returns:
    --------
        The bokeh object
    """
    colors = [
        "#75968f",
        "#a5bab7",
        "#c9d9d3",
        "#e2e2e2",
        "#dfccce",
        "#ddb7b1",
        "#cc7878",
        "#933b41",
        "#550b1d",
    ]
    colors = RdBu[8]
    mapper = LinearColorMapper(
        palette=colors, low=df.Segment_Mean.min(), high=df.Segment_Mean.max()
    )
    if len(sample_order) == 0:
        sample_order = list(set(df.Sample.tolist()))
    TOOLS = "hover,save,pan,box_zoom,reset,wheel_zoom"
    p = figure(
        title=title,
        y_range=(df.End.max(), df.Start.min()),
        x_range=sample_order,
        x_axis_location="above",
        plot_width=width,
        plot_height=height,
        tools=TOOLS,
        toolbar_location="below",
        tooltips=[("pos", "@Start, @End"), ("relative CN", "@Sample")],
    )

    p.grid.grid_line_color = None
    p.axis.axis_line_color = None
    p.axis.major_tick_line_color = None
    p.axis.major_label_text_font_size = "5pt"
    p.axis.major_label_standoff = standoff
    p.xaxis.major_label_orientation = pi / 3
    pos = 0
    # for i,val in enumerate(historder):
    #    p.rect(x=pos,y=-7,width=len(orderbyhist[val]), height=10, fill_color=small_palettes['Viridis'][4][i])
    #    p.text(x=pos+len(orderbyhist[val])/2, y=-9, text=str(val), text_color="#96deb3")
    #    pos += len(orderbyhist[val])

    p.rect(
        x="Sample",
        y="Start",
        width=0.9,
        height="size",
        source=df.reset_index(drop=True),
        fill_color={"field": "Segment_Mean", "transform": mapper},
        line_color=None,
    )

    color_bar = ColorBar(
        color_mapper=mapper,
        major_label_text_font_size="5pt",
        ticker=BasicTicker(desired_num_ticks=len(colors)),
        formatter=PrintfTickFormatter(format="%.2f"),
        label_standoff=6,
        border_line_color=None,
        location=(0, 0),
    )
    p.add_layout(color_bar, "right")
    p.yaxis.axis_label = y_label
    # p.yaxis.major_label_overrides={20:'Centromer'}
    for val in marks:
        hline = Span(
            location=val, dimension="width", line_color="green", line_width=0.2
        )
        p.renderers.extend([hline])
    p.output_backend = "svg"
    if folder:
        save(p, folder + title.replace(" ", "_") + "_cn_plot.html")
        try:
            export_svg(p, filename=folder + title.replace(" ", "_") + "_cn_plot.svg")
        except RuntimeError:
            print("You need to install Selenium to save the svg!")
    show(p)  # show the plot
    return p


def volcano(
    data,
    folder="",
    tohighlight=None,
    tooltips=[("gene", "@gene_id")],
    title="volcano plot",
    xlabel="log-fold change",
    ylabel="-log(Q)",
    maxvalue=100,
    searchbox=False,
    logfoldtohighlight=0.15,
    pvaltohighlight=0.1,
    showlabels=False,
):
    """
    Make an interactive volcano plot from Differential Expression analysis tools outputs

    Args:
    -----
        data: a df with rows genes and cols [log2FoldChange, pvalue, gene_id]
        folder: str of location where to save the plot, won't save if empty
        tohighlight: list[str] of genes to highlight in the plot
        tooltips: list[tuples(str,str)] if user wants tot specify another bokeh tooltip
        title: str plot title
        xlabel: str if user wants to specify the title of the x axis
        ylabel: str if user wants tot specify the title of the y axis
        maxvalue: float the max -log2(pvalue authorized usefull when managing inf vals)
        searchbox: bool whether or not to add a searchBox to interactively highlight genes
        logfoldtohighlight: float min logfoldchange when to diplay points
        pvaltohighlight: float min pvalue when to diplay points
        showlabels: bool whether or not to show a text above each datapoint with its label information

    Returns:
    --------
        The bokeh object
    """
    # pdb.set_trace()
    to_plot_not, to_plot_yes = selector(
        data,
        tohighlight if tohighlight is not None else [],
        logfoldtohighlight,
        pvaltohighlight,
    )
    hover = bokeh.models.HoverTool(tooltips=tooltips, names=["circles"])

    # Create figure
    p = bokeh.plotting.figure(title=title, plot_width=650, plot_height=450)

    p.xgrid.grid_line_color = "white"
    p.ygrid.grid_line_color = "white"
    p.xaxis.axis_label = xlabel
    p.yaxis.axis_label = ylabel

    # Add the hover tool
    p.add_tools(hover)
    p, source1 = add_points(
        p, to_plot_not, "log2FoldChange", "pvalue", color="#1a9641", maxvalue=maxvalue
    )
    p, source2 = add_points(
        p,
        to_plot_yes,
        "log2FoldChange",
        "pvalue",
        color="#fc8d59",
        alpha=0.6,
        outline=True,
        maxvalue=maxvalue,
    )
    if showlabels:
        labels = LabelSet(
            x="log2FoldChange",
            y="transformed_q",
            text_font_size="7pt",
            text="gene_id",
            level="glyph",
            x_offset=5,
            y_offset=5,
            source=source2,
            render_mode="canvas",
        )
        p.add_layout(labels)
    if searchbox:
        text = TextInput(title="text", value="gene")
        text.js_on_change(
            "value",
            CustomJS(
                args=dict(source=source1),
                code="""
                var data = source.data
                var value = cb_obj.value
                var gene_id = data.gene_id
                var a = -1
                for (let i=0; i < gene_id.length; i++) {
                    if ( gene_id[i]===value ) { a=i; console.log(i); data.size[i]=7; data.alpha[i]=1; data.color[i]='#fc8d59' }
                }
                source.data = data
                console.log(source)
                console.log(cb_obj)
                source.change.emit()
                console.log(source)
                """,
            ),
        )
        p = column(text, p)
    p.output_backend = "svg"
    if folder:
        save(p, folder + title.replace(" ", "_") + "_volcano.html")
        try:
            export_svg(p, filename=folder + title.replace(" ", "_") + "_volcano.svg")
        except RuntimeError:
            print("You need to install Selenium to save the svg!")
    try:
        show(p)
    except:
        show(p)
    return p


def add_points(p, df1, x, y, color="blue", alpha=0.2, outline=False, maxvalue=100):
    """parts of volcano plot"""
    # Define colors in a dictionary to access them with
    # the key from the pandas groupby funciton.
    df = df1.copy()
    transformed_q = -df[y].apply(np.log10).values
    transformed_q[transformed_q == np.inf] = maxvalue
    df["transformed_q"] = transformed_q
    df["color"] = color
    df["alpha"] = alpha
    df["size"] = 7
    source1 = bokeh.models.ColumnDataSource(df)

    # Specify data source
    p.scatter(
        x=x,
        y="transformed_q",
        size="size",
        alpha="alpha",
        source=source1,
        color="color",
        name="circles",
    )
    if outline:
        p.scatter(
            x=x,
            y="transformed_q",
            size=7,
            alpha=1,
            source=source1,
            color="black",
            fill_color=None,
            name="outlines",
        )

    # prettify
    p.background_fill_color = "#DFDFE5"
    p.background_fill_alpha = 0.5
    return p, source1


def selector(
    df,
    valtoextract=[],
    logfoldtohighlight=0.15,
    pvaltohighlight=0.1,
    minlogfold=0.15,
    minpval=0.1,
):
    """Part of Volcano plot: A function to separate tfs from everything else"""
    toshow = (df.pvalue < minpval) & (abs(df.log2FoldChange) > minlogfold)
    df = df[toshow]
    sig = (df.pvalue < pvaltohighlight) & (abs(df.log2FoldChange) > logfoldtohighlight)
    if valtoextract:
        not_tf = ~df.gene_id.isin(valtoextract)
        is_tf = df.gene_id.isin(valtoextract)
        to_plot_not = df[~sig | not_tf]
        to_plot_yes = df[sig & is_tf]
    else:
        to_plot_not = df[~sig]
        to_plot_yes = df[sig]
    return to_plot_not, to_plot_yes


# What pops up on hover?


def correlationMatrix(
    data,
    names,
    colors=None,
    pvals=None,
    maxokpval=10**-9,
    other=None,
    title="correlation Matrix",
    dataIsCorr=False,
    invert=False,
    size=40,
    folder="",
    interactive=False,
    maxval=None,
    minval=None,
):
    """
    Make an interactive correlation matrix from an array using bokeh

    Args:
    -----
      data: arrayLike of int / float/ bool of size(names*val) or (names*names)
      names: list[str] of names for each rows
      colors: list[int] of size(names) a color for each names (good to display clusters)
      pvals: arraylike of int / float/ bool of size(names*val) or (names*names) with the corresponding pvalues
      maxokpval: float threshold when pvalue is considered good. otherwise lowers the size of the square
        until 10**-3 when it disappears
      other: arrayLike of int / float/ bool of size(names*val) or (names*names), an additional information
        matrix that you want ot display with opacity whereas correlations willl be displayed with
      title: str the plot title
      dataIsCorr: bool if not true, we will compute the corrcoef of the data array
      invert: bool whether or not to invert the matrix before running corrcoef
      size: int the plot size
      folder: str of folder location where to save the plot, won't save if empty
      interactive: bool whether or not to make the plot interactive (else will use matplotlib)
      maxval: float clamping coloring up to maxval
      minval: float clamping coloring down to minval

    Returns:
    -------
      the bokeh object if interactive else None

    """
    if not dataIsCorr:
        print("computing correlations")
        data = np.corrcoef(np.array(data) if not invert else np.array(data).T)
    else:
        data = np.array(data)
    regdata = data.copy()
    if minval is not None:
        data[data < minval] = minval
    if maxval is not None:
        data[data > maxval] = maxval
    data = data / data.max()
    TOOLS = (
        "hover,crosshair,pan,wheel_zoom,zoom_in,zoom_out,box_zoom,undo,redo,reset,save"
    )
    xname = []
    yname = []
    color = []
    alpha = []
    height = []
    width = []
    if type(colors) is list:
        print("we are assuming you want to display clusters with colors")
    elif other is not None:
        print(
            "we are assuming you want to display the other of your correlation with opacity"
        )
    if pvals is not None:
        print(
            "we are assuming you want to display the pvals of your correlation with size"
        )
        regpvals = pvals.copy()
        u = pvals < maxokpval
        pvals[~u] = np.log10(1 / pvals[~u])
        pvals = pvals / pvals.max()
        pvals[u] = 1
    if interactive:
        xname = []
        yname = []
        color = []
        for i, name1 in enumerate(names):
            for j, name2 in enumerate(names):
                xname.append(name1)
                yname.append(name2)
                if pvals is not None:
                    height.append(max(0.1, min(0.9, pvals[i, j])))
                    color.append(cc.coolwarm[int((data[i, j] * 127) + 127)])
                    alpha.append(min(abs(data[i, j]), 0.9))
                elif other is not None:
                    color.append(cc.coolwarm[int((data[i, j] * 127) + 127)])
                    alpha.append(
                        max(min(other[i, j], 0.9), 0.1) if other[i, j] != 0 else 0
                    )
                else:
                    alpha.append(min(abs(data[i, j]), 0.9))
                if colors is not None:
                    if type(colors) is list:
                        if colors[i] == colors[j]:
                            color.append(Category10[10][colors[i]])
                        else:
                            color.append("lightgrey")

                elif pvals is None and other is None:
                    color.append("grey" if data[i, j] > 0 else Category20[3][2])
        print(regdata.max())
        if pvals is not None:
            width = height.copy()
            data = dict(
                xname=xname,
                yname=yname,
                colors=color,
                alphas=alpha,
                data=regdata.ravel(),
                pvals=regpvals.ravel(),
                width=width,
                height=height,
            )
        else:
            data = dict(
                xname=xname, yname=yname, colors=color, alphas=alpha, data=data.ravel()
            )
        tt = [("names", "@yname, @xname"), ("value", "@data")]
        if pvals is not None:
            tt.append(("pvals", "@pvals"))
        p = figure(
            title=title if title is not None else "Correlation Matrix",
            x_axis_location="above",
            tools=TOOLS,
            x_range=list(reversed(names)),
            y_range=names,
            tooltips=tt,
        )

        p.plot_width = 800
        p.plot_height = 800
        p.grid.grid_line_color = None
        p.axis.axis_line_color = None
        p.axis.major_tick_line_color = None
        p.axis.major_label_text_font_size = "5pt"
        p.axis.major_label_standoff = 0
        p.xaxis.major_label_orientation = np.pi / 3
        p.output_backend = "svg"
        p.rect(
            "xname",
            "yname",
            width=0.9 if not width else "width",
            height=0.9 if not height else "height",
            source=data,
            color="colors",
            alpha="alphas",
            line_color=None,
            hover_line_color="black",
            hover_color="colors",
        )
        save(p, folder + title.replace(" ", "_") + "_correlation.html")
        try:
            export_svg(
                p, filename=folder + title.replace(" ", "_") + "_correlation.svg"
            )
        except RuntimeError:
            print("You need to install Selenium to save the svg!")
        try:
            show(p)
        except:
            show(p)
        return p  # show the plot
    else:
        plt.figure(figsize=(size, 200))
        plt.title("the correlation matrix")
        plt.imshow(data)
        plt.savefig(title + "_correlation.pdf")
        plt.show()


def heatmap(
    data,
    colors=None,
    title="correlation Matrix",
    size=40,
    folder="",
    interactive=False,
    pvals=None,
    maxokpval=10**-9,
    maxval=None,
    minval=None,
):
    """
    Make an interactive heatmap from a dataframe using bokeh

    Args:
    -----
      data: dataframe of int / float/ bool of size(names1*names2)
      colors: list[int] of size(names) a color for each names (good to display clusters)
      pvals: arraylike of int / float/ bool of size(names*val) or (names*names) with the corresponding pvalues
      maxokpval: float threshold when pvalue is considered good. otherwise lowers the size of the square
        until 10**-3 when it disappears
      title: str the plot title
      size: int the plot size
      folder: str of folder location where to save the plot, won't save if empty
      interactive: bool whether or not to make the plot interactive (else will use matplotlib)
      maxval: float clamping coloring up to maxval
      minval: float clamping coloring down to minval

    Returns:
    -------
      the bokeh object if interactive else None

    """
    regdata = data.copy()
    if minval is not None:
        data[data < minval] = minval
    if maxval is not None:
        data[data > maxval] = maxval
    data = data / data.max()
    data = data.values
    TOOLS = (
        "hover,crosshair,pan,wheel_zoom,zoom_in,zoom_out,box_zoom,undo,redo,reset,save"
    )
    xname = []
    yname = []
    color = []
    alpha = []
    height = []
    width = []
    if pvals is not None:
        print(
            "we are assuming you want to display the pvals of your correlation with size"
        )
        regpvals = pvals.copy()
        u = pvals < maxokpval
        pvals[~u] = np.log10(1 / pvals[~u])
        pvals = pvals / pvals.max()
        pvals[u] = 1
    if interactive:
        xname = []
        yname = []
        color = []
        for i, name1 in enumerate(regdata.index):
            for j, name2 in enumerate(regdata.columns):
                xname.append(name2)
                yname.append(name1)
                if pvals is not None:
                    # import pdb;pdb.set_trace()
                    height.append(max(0.1, min(0.9, pvals.loc[name1][name2])))
                    color.append(cc.coolwarm[int((data[i, j] * 128) + 127)])
                    alpha.append(min(abs(data[i, j]), 0.9))
                elif other is not None:
                    color.append(cc.coolwarm[int((data[i, j] * 128) + 127)])
                    alpha.append(
                        max(min(other[i, j], 0.9), 0.1) if other[i, j] != 0 else 0
                    )
                else:
                    alpha.append(min(abs(data[i, j]), 0.9))
                if colors is not None:
                    if type(colors) is list:
                        if colors[i] == colors[j]:
                            color.append(Category10[10][colors[i]])
                        else:
                            color.append("lightgrey")

                elif pvals is None and other is None:
                    color.append("grey" if data[i, j] > 0 else Category20[3][2])
        if pvals is not None:
            width = height.copy()
            data = dict(
                xname=xname,
                yname=yname,
                colors=color,
                alphas=alpha,
                data=regdata.values.ravel(),
                pvals=regpvals.values.ravel(),
                width=width,
                height=height,
            )
        else:
            data = dict(
                xname=xname, yname=yname, colors=color, alphas=alpha, data=data.ravel()
            )
        tt = [("names", "@yname, @xname"), ("value", "@data")]
        if pvals is not None:
            tt.append(("pvals", "@pvals"))
        p = figure(
            title=title if title is not None else "Heatmap",
            x_axis_location="above",
            tools=TOOLS,
            x_range=list(reversed(regdata.columns.astype(str).tolist())),
            y_range=regdata.index.tolist(),
            tooltips=tt,
        )

        p.plot_width = 800
        p.plot_height = 800
        p.grid.grid_line_color = None
        p.axis.axis_line_color = None
        p.axis.major_tick_line_color = None
        p.axis.major_label_text_font_size = "5pt"
        p.axis.major_label_standoff = 0
        p.xaxis.major_label_orientation = np.pi / 3
        p.output_backend = "svg"
        p.rect(
            "xname",
            "yname",
            width=0.9 if not width else "width",
            height=0.9 if not height else "height",
            source=data,
            color="colors",
            alpha="alphas",
            line_color=None,
            hover_line_color="black",
            hover_color="colors",
        )
        save(p, folder + title.replace(" ", "_") + "_heatmap.html")
        try:
            export_svg(
                p, filename=folder + title.replace(" ", "_") + "_correlation.svg"
            )
        except RuntimeError:
            print("You need to install Selenium to save the svg!")
        try:
            show(p)
        except:
            show(p)
        return p  # show the plot
    else:
        plt.figure(figsize=size)
        plt.title("the correlation matrix")
        plt.imshow(data)
        plt.savefig(title + "_correlation.pdf")
        plt.show()


def venn(inp, names, title="venn", folder=""):
    """
    Plots a venn diagram using the pyvenn package

    Args:
    -----
        inp: list[set()] of sets of values (e.g. [(1,2,3,4),(2,3),(1,3,4,5)])
        names: list[str] of the name of each leaf
        title: str the plot title
        folder: str of location where to save the plot, won't save if empty
    """
    labels = pyvenn.get_labels(inp, fill=["number", "logic"])
    if len(inp) == 2:
        fig, ax = pyvenn.venn2(labels, names=names)
    elif len(inp) == 3:
        fig, ax = pyvenn.venn3(labels, names=names)
    elif len(inp) == 4:
        fig, ax = pyvenn.venn4(labels, names=names)
    elif len(inp) == 5:
        fig, ax = pyvenn.venn5(labels, names=names)
    elif len(inp) == 6:
        fig, ax = pyvenn.venn6(labels, names=names)
    else:
        raise ValueError("need to be between 2 to 6")
    ax.set_title(title)
    if folder:
        fig.savefig(folder + title + "_venn.pdf")
    fig.show()
    plt.pause(0.1)


def mergeImages(images, outputpath):
    """
    will merge a set of images in python

    Args:
    -----
        images: list of image filepath
        outputpath: where to save the resulting merger
    """
    images = list(map(Image.open, images))
    widths, heights = zip(*(i.size for i in images))

    total_width = max(widths)
    max_height = sum(heights)

    new_im = Image.new("RGB", (total_width, max_height))

    y_offset = 0
    for im in images:
        new_im.paste(im, (0, y_offset))
        y_offset += im.size[1]

    new_im.save(outputpath)


def addTextToImage(image, text, outputpath, xy=(0, 0), color=(0, 0, 0), fontSize=64):
    """
    will add some text to an image in python

    Args:
    ----
        image: the image filepath
        text: the text to write
        outputpath: the location of the resulting image
        xy: the location of the text
        color: tuple(a,b,c) a tuple of 3 ints between 0 and 256
        fontSize: an int for the font size
    """
    # adds black text to the upper left by default, Arial size 64
    img = Image.open(image)
    draw = ImageDraw.Draw(img)
    # the below file path assumes you're operating macOS
    font = ImageFont.truetype("/Library/Fonts/Arial.ttf", fontSize)
    draw.text(xy, text, color, font=font)
    img.save(outputpath)


def SOMPlot(
    net, size, colnames, minweight=0.1, distq1=0.535, distq2=0.055, distr=0.2, folder=""
):
    """
    makes an interactive plot from a SOM from the SimpSOM package

    a tool that uses simpSOM's package output (which produces self organizing maps),
    to plot its output in an interactive fashion

    Args:
    -----
      net: SIMPSOM's net object, output from somplot
      size: float the size of the plot
      distq1: float a value to adjust to considering the numbe of nodes (controls spacing beween cols of nodes)
      distq2: float a value to adjust to considering the numbe of nodes (controls spacing beween rows of nodes)
      distr: float a value to adjust to considering the numbe of nodes (controls nodes size spacing)
      folder: str of location where to save the plot, won't save if empty
    """
    diffs = net.diff_graph(show=False, returns=True)
    somnodes = {"r": [], "q": [], "c": diffs, "features": []}
    for i, node in enumerate(net.nodeList):
        somnodes["q"].append(node.pos[0] + (i % size) * distq1 + (i // size) * distq2)
        somnodes["r"].append(-node.pos[1] - (i % size) * distr)
        somnodes["features"].append(
            [
                colnames[i]
                for i in np.argsort(node.weights)
                if abs(node.weights[i]) > minweight
            ]
        )
    somnodes = pd.DataFrame(somnodes)
    for i, v in somnodes.iterrows():
        tot = ""
        for e, j in enumerate(v.features):
            if e % 5 == 4:
                tot += "\n"
            tot += " " + str(j)
        somnodes.loc[i, "features"] = tot
    # interactive SOM with features with highest importance to the nodes, displayed when hovering
    bigScatter(
        somnodes,
        precomputed=True,
        features=True,
        binsize=1,
        title="Cobinding SOM cluster of " + str(size),
        folder=folder,
    )


def quartdist(x):
    """
    creates a dist plot from a dataframe with quartiles

    Args:
    -----
        x: dataframe to run kdeplot on

    """
    ax, _ = plt.subplots(figsize=(10, 10))
    sns.kdeplot(x, shade=False, color="crimson", ax=ax)
    kdeline = ax.lines[0]
    xs = kdeline.get_xdata()
    ys = kdeline.get_ydata()
    left, middle, right = np.percentile(x, [25, 50, 75])
    ax.set_title("Showing median and quartiles")
    ax.vlines(middle, 0, np.interp(middle, xs, ys), color="crimson", ls=":")
    ax.fill_between(xs, 0, ys, facecolor="crimson", alpha=0.2)
    ax.fill_between(
        xs,
        0,
        ys,
        where=(left <= xs) & (xs <= right),
        interpolate=True,
        facecolor="crimson",
        alpha=0.2,
    )
    ax.set_ylim(ymin=0)
    plt.show()



import gspread
from oauth2client.service_account import ServiceAccountCredentials

scope = ["https://spreadsheets.google.com/feeds", 'https://www.googleapis.com/auth/spreadsheets',
         "https://www.googleapis.com/auth/drive.file", "https://www.googleapis.com/auth/drive"]


def dfToSheet(df, sheetid, secret='~/.credentials.json'):
  credentials = ServiceAccountCredentials.from_json_keyfile_name(secret, scope)
  client = gspread.authorize(credentials)
  spreadsheet = client.open(sheetid)
  df.to_csv('/tmp/sheet.csv')
  with open("/tmp/sheet.csv", 'r') as file_obj:
    content = file_obj.read()
    client.import_csv(spreadsheet.id, data=content)



from setuptools import setup
import sys
import os
import io
import subprocess

if sys.version_info.major < 3 or sys.version_info.minor < 2:
    raise ValueError("genepy is only compatible with Python 3.5 and above")
if sys.version_info.minor < 5:
    import warnings

    warnings.warn("genepy may not function properly on Python < 3.8")

os.system("git submodule init && git submodule sync")

with open("README.md", "r") as f:
    long_description = f.read()

print("trying to install R packages")
try:
    subprocess.run(
        'R -e \'if(!requireNamespace("BiocManager", quietly = TRUE)){install.packages("BiocManager", repos="http://cran.us.r-project.org")};BiocManager::install(c("GSEABase", "erccdashboard", "GSVA", "DESeq2"));\'',
        shell=True,
        check=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    subprocess.run("pip install rpy2")
except:
    print("R packages not installed")
print("if it did not work. please install R or check your R installation")
print(
    "once R is installed you need to install erccdashboard, GSEABase GSVA, DESeq2 to have access to all the functions"
)


def read(*paths, **kwargs):
    """Read the contents of a text file safely.
    >>> read("genepy", "VERSION")
    '0.1.0'
    >>> read("README.md")
    ...
    """

    content = ""
    with io.open(
        os.path.join(os.path.dirname(__file__), *paths),
        encoding=kwargs.get("encoding", "utf8"),
    ) as open_file:
        content = open_file.read().strip()
    return content


def read_requirements(path):
    return [
        line.strip()
        for line in read(path).split("\n")
        if not line.startswith(('"', "#", "-", "git+"))
    ]


setup(
    name="Broad-genepy",
    version=read("genepy", "VERSION"),
    description="A useful module for any CompBio",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Jeremie Kalfon",
    author_email="jkobject@gmail.com",
    url="https://github.com/BroadInstitute/genepy",
    packages=[
        "genepy/cell_line_mapping-master/python/cell_line_mapper",
        "genepy/epigenetics",
        "genepy/mutations",
        "genepy/google",
        "genepy/sequencing/",
        "genepy/terra",
        "genepy/rna",
        "genepy/utils",
    ],
    package_data={"genepy": ["data/*"]},
    python_requires=">=3.5",
    install_requires=read_requirements("requirements.txt"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
)


print(
    "You might want to install Bowtie2, samtools, bwa and R to be able to use all functions of this package:\n\
  http://bowtie-bio.sourceforge.net/bowtie2/index.shtml\n\
  http://www.htslib.org/\n\
  https://github.com/lh3/bwa\n"
)

print("Finished!")



# Jeremie Kalfon
# for BroadInsitute
# in 2019

from __future__ import print_function
import json
import os
import string
import subprocess
import pandas as pd
import numpy as np
import itertools
import random
import xmltodict
from biomart import BiomartServer
import io

chromsize = {
  "chr1":	248956422,
  "chr10":133797422,
  "chr11":135086622,
  "chr12":133275309,
  "chr13":114364328,
  "chr14":107043718,
  "chr15":101991189,
  "chr16":90338345,
  "chr17":83257441,
  "chr18":80373285,
  "chr19":58617616,
  "chr2":	242193529,
  "chr20":64444167,
  "chr21":46709983,
  "chr22":50818468,
  "chr3":	198295559,
  "chr4":	190214555,
  "chr5":	181538259,
  "chr6":	170805979,
  "chr7":	159345973,
  "chr8":	145138636,
  "chr9":	138394717,
  "chrX":	156040895,
  "chrY":	57227415,
}

rename_mut = {'contig': 'chr', 'position': 'pos', 'Reference_Allele': 'ref', 'ref_allele': 'ref', 'alt_allele': 'alt',
            'Chromosome': 'chr', 'End_postition': 'end', 'Start_position': 'pos', 'Tumor_Seq_Allele1': "alt"}


def fileToList(filename):
  """
  loads an input file with a\\n b\\n.. into a list [a,b,..]
  """
  with open(filename) as f:
    return [val[:-1] for val in f.readlines()]


def listToFile(l, filename):
  """
  loads a list with [a,b,..] into an input file a\\n b\\n..
  """
  with open(filename, 'w') as f:
    for item in l:
      f.write("%s\n" % item)


def dictToFile(d, filename):
  """
  turn a dict into a json file
  """
  with open(filename, 'w') as json_file:
    json.dump(d, json_file)


def fileToDict(filename):
  """
  loads a json file into a python dict
  """
  with open(filename) as f:
    data = json.load(f)
  return data


def batchMove(l, pattern=['*.', '.*'], folder='', add=''):
  """
  moves a set of files l into a folder:

  Args:
    l: file list
    pattern: if files are a set of patterns to match
    folder: folder to move file into
    add: some additional mv parameters
  """
  for val in l:
    cmd = 'mv '
    if add:
      cmd += add + ' '
    if '*.' in pattern:
      cmd += '*'
    cmd += val
    if '.*' in pattern:
      cmd += '*'
    cmd += " " + folder
    res = os.system(cmd)
    if res != 0:
      raise Exception("Leave command pressed or command failed")


def batchRename(dt, folder='', sudo=False, doAll=False, add='', dryrun=False):
  """
  Given a dict renames corresponding files in a folder

  Args:
    dt (dict): dict(currentName:newName) renaming dictionnary
    folder (str): folder to look into
    add: some additional mv parameters
  """
  cmd = 'ls -R ' + folder if doAll else 'ls ' + folder
  files = os.popen(cmd).read().split('\n')
  if doAll:
    prep=''
    f = []
    for val in files:
      if len(val)==0:
        prep=''
        continue
      if val[0]=='.' and len(val)>3:
        prep=val[:-1]
        continue
      if "." in val:
        f.append(prep+"/"+val)
    files = f
  for k, val in dt.items():
    for f in files:
      if k in f:
        cmd = 'sudo mv ' if sudo else 'mv '
        if add:
          cmd += add + ' '
        if not doAll:
          cmd += folder
        cmd += f
        cmd += ' '
        if not doAll:
          cmd += folder
        cmd += f.replace(k, val)
        if dryrun:
          print(cmd)
        else:
          res = os.system(cmd)
          if res != 0:
            raise Exception("Leave command pressed or command failed")


def grouped(iterable, n):
  """
  iterate over element of list 2 at a time python

  s -> (s0,s1,s2,...sn-1), (sn,sn+1,sn+2,...s2n-1), (s2n,s2n+1,s2n+2,...s3n-1), ...
  """
  it = iter(iterable)
  while True:
    chunk = tuple(itertools.islice(it, n))
    if not chunk:
      return
    yield chunk


def overlap(interval1, interval2):
  """computed overlap

  Given [0, 4] and [1, 10] returns [1, 4]
  Given [0, 4] and [8, 10] returns False
  """
  if interval2[0] <= interval1[0] <= interval2[1]:
    start = interval1[0]
  elif interval1[0] <= interval2[0] <= interval1[1]:
    start = interval2[0]
  else:
    return False

  if interval2[0] <= interval1[1] <= interval2[1]:
    end = interval1[1]
  elif interval1[0] <= interval2[1] <= interval1[1]:
    end = interval2[1]
  else:
    return False

  return (start, end)


def union(interval1, interval2):
  """
  Given [0, 4] and [1, 10] returns [0, 10]
  Given [0, 4] and [8, 10] returns False
  """
  if interval1[0] <= interval2[0] <= interval1[1]:
    start = interval1[0]
    end = interval1[1] if interval2[1] <= interval1[1] else interval2[1]
  elif interval1[0] <= interval2[1] <= interval1[1]:
    start = interval2[0] if interval2[0] <= interval1[0] else interval1[0]
    end = interval1[1]
  else:
    return False
  return (start, end)


def nans(df): return df[df.isnull().any(axis=1)]


def createFoldersFor(filepath):
  """
  will recursively create folders if needed until having all the folders required to save the file in this filepath
  """
  prevval = ''
  for val in os.path.expanduser(filepath).split('/')[:-1]:
    prevval += val + '/'
    if not os.path.exists(prevval):
      os.mkdir(prevval)


def randomString(stringLength=6, stype='all', withdigits=True):
  """
  Generate a random string of letters and digits

  Args:
    stringLength (int, optional): the amount of char. Defaults to 6.
    stype (str, optional): one of lowercase, uppercase, all. Defaults to 'all'.
    withdigits (bool, optional): digits allowed in the string? Defaults to True.

  Returns:
    str: random string
  """
  if stype == 'lowercase':
    lettersAndDigits = string.ascii_lowercase
  elif stype == 'uppercase':
    lettersAndDigits = string.ascii_uppercase
  else:
    lettersAndDigits = string.ascii_letters
  if withdigits:
    lettersAndDigits += string.digits
  return ''.join(random.choice(lettersAndDigits) for i in range(stringLength))


def pdDo(df, op="mean", of="value1", over="value2"):
  """
  apply a function to a panda dataframe WIP
  """
  df = df.sort_values(by=over)
  index = []
  data = df.iloc[0, of]
  ret = []
  prev = df.iloc[0, over]
  j = 0
  for k, val in df.iloc[1:].iterrows():
    if val[over] == prev:
      data.append(val[of])
    else:
      if of == "mean":
        ret[j] = np.mean(data)
      elif of == "sum":
        ret[j] = np.sum(data)
      elif of == "max":
        ret[j] = np.max(data)
      elif of == "min":
        ret[j] = np.min(data)
      index.append(k)
      j += 1
      data = [val[of]]
  return index, ret


def parrun(cmds, cores, add=[]):
  """
  runs a set of commands in parallel using the "&" command

  Args:
    cmds: the list of commands
    cores: number of parallel execution
    add: an additional list(len(cmds)) of command to run in parallel at the end of each parallel run
  """
  count = 0
  exe = ''
  if len(add) != 0 and len(add) != len(cmds):
    raise ValueError("we would want them to be the same size")
  else:
    addexe = ''
  fullres = []
  for i, cmd in enumerate(cmds):
    count += 1
    exe += cmd
    if len(add) != 0:
      addexe += add[i]
    if count < cores and i < len(cmds) - 1:
      exe += ' & '
      if len(add) != 0:
        addexe += ' & '
    else:
      count = 0
      res = subprocess.run(exe, capture_output=True, shell=True)
      if res.returncode != 0:
        raise ValueError('issue with the command: ' + str(res.stderr))
      exe = ''
      if len(add) != 0:
        res = subprocess.run(addexe, capture_output=True, shell=True)
        if res.returncode != 0:
          raise ValueError(
            'issue with the command: ' + str(res.stderr))
        addexe = ''
      fullres.append(res.stdout.decode('utf-8'))
  return fullres


def askif(quest):
    """
    asks a y/n question to the user about something and returns true or false given his answer
    """
    print(quest)
    inp = input()
    if inp in ['yes', 'y', 'Y', 'YES', 'oui', 'si']:
      return 1
    elif inp in ['n', 'no', 'nope', 'non', 'N']:
      return 0
    else:
      return askif('you need to answer by yes or no')


def inttodate(i, lim=1965, unknown='U', sep='-', order="asc", startsatyear=0):
    """
    transforms an int representing days into a date

    Args:
      i: the int
      lim: the limited year below which we have a mistake
      unknown: what to return when unknown (date is bellow the limited year)
      sep: the sep between your date (e.g. /, -, ...)
      order: if 'asc', do d,m,y else do y,m,d
      startsatyear: when is the year to start counting for this int

    Returns:
      str: the date or unknown
    """
    a = int(i // 365)
    if a > lim:
        a = str(a + startsatyear)
        r = i % 365
        m = str(int(r // 32)) if int(r // 32) > 0 else str(1)
        r = r % 32
        d = str(int(r)) if int(r) > 0 else str(1)
    else:
        return unknown
    return d + sep + m + sep + a if order == "asc" else a + sep + m + sep + d


def datetoint(dt, split='-', unknown='U', order="des"):
  """
  same as inttodate but in the opposite way;

  starts at 0y,0m,0d

  Args:
    dt: the date string
    split: the splitter in the string (e.g. /,-,...)
    unknown: maybe the some dates are 'U' or 0 and the program will output 0 for unknown instead of crashing
    order: if 'asc', do d,m,y else do y,m,d

  Returns:
    int: the date
  """
  arr = np.array(dt[0].split(split) if dt[0] !=
                  unknown else [0, 0, 0]).astype(int)
  if len(dt) > 1:
    for val in dt[1:]:
      arr = np.vstack(
        (arr, np.array(val.split(split) if val != unknown and val.count(split) == 2 else [0, 0, 0]).astype(int)))
    arr = arr.T
  res = arr[2] * 365 + arr[1] * 31 + \
    arr[0] if order == "asc" else arr[0] * 365 + arr[1] * 31 + arr[2]
  return [res] if type(res) is np.int64 else res

prevshowcount = 100

def showcount(i, size):
    """
    pretty print of i/size%, to put in a for loop
    """
    global prevshowcount
    a = 1 + int(100 * (i / size))
    if a != prevshowcount:
        print(str(a) + '%', end='\r')
        prevshowcount = a


def combin(n, k):
  """
  Nombre de combinaisons de n objets pris k a k
  outputs the number of comabination of n object taken k at a time
  """
  if k > n // 2:
      k = n - k
  x = 1
  y = 1
  i = n - k + 1
  while i <= n:
    x = (x * i) // y
    y += 1
    i += 1
  return x


def dups(lst):
  """
      shows the duplicates in a list
  """
  seen = set()
  # adds all elements it doesn't know yet to seen and all other to seen_twice
  seen_twice = set(x for x in lst if x in seen or seen.add(x))
  # turn the set into a list (as requested)
  return list(seen_twice)


def makeCombinations(size, proba):
  """
  produces probability of X event happening at the same time

  pretty usefull for cobinding analysis. wil compute it
  given binomial probabilities of each event occuring and the number of trials

  Args:
    size: int number of trials
    proba: list[float] probabilities of each event occuring
  """
  sums = {i:0 for i in range(1,size)}
  for i in range(size-1,0,-1):
    print(i)
    if sums[i]> 0:
      continue
    print(combin(size+3,i))
    v=0
    for j in itertools.combinations(proba, i):
      v+=np.prod(j)
    sums[i] = v
  for i in range(size-1,0,-1):
    for j in range(i+1,size):
      icomb = combin(j,i)
      sums[i] -= icomb*sums[j]
  sums[0] = 1-sum(list(sums.values()))
  return sums


def closest(lst, K):
  """
  returns the index of the value closest to K in a lst
  """
  return lst[min(range(len(lst)), key = lambda i: abs(lst[i]-K))]


def compareDfs(df1, df2):
  """
  compares df1 to df2

  shows col difference, index difference, nans & 0s differences
  """
  nmissmatchCols = set(df1.columns)-set(df2.columns)
  omissmatchCols = set(df2.columns)-set(df1.columns)
  nmissmatchInds = set(df1.index)-set(df2.index)
  omissmatchInds = set(df2.index)-set(df1.index)
  newNAs = df1.isna().sum().sum() - df2.isna().sum().sum()
  new0s = (df1 == 0).sum().sum() - (df2 == 0).sum().sum()
  print('FOUND missmatch Columns NOT IN df2: ' + str(nmissmatchCols))
  print('FOUND missmatch Columns NOT IN df1: ' + str(omissmatchCols))
  print('FOUND missmatch Index NOT IN df2: ' + str(nmissmatchInds))
  print('FOUND missmatch Index NOT IN df1: ' + str(omissmatchInds))
  print('FOUND new NAs in df1: ' + str(newNAs))
  print('FOUND new 0s in df1: ' + str(new0s))
  return nmissmatchCols, omissmatchCols, nmissmatchInds, omissmatchInds, newNAs, new0s


def stringifydict(res):
  """[summary]

  Args:
      res ([type]): [description]

  Returns:
      [type]: [description]
  """
  a = {}
  for k, v in res.items():
    if type(v) is dict:
      a[k] = stringifydict(v)
    else:
      a[str(k)] = v
  return a


def readXMLs(folder=None, file=None, rename=None):
  """[summary]

  Args:
      folder ([type], optional): [description]. Defaults to None.
      file ([type], optional): [description]. Defaults to None.
      rename ([type], optional): [description]. Defaults to None.

  Raises:
      ValueError: [description]

  Returns:
      [type]: [description]
  """
  if file is not None:
    if type(file) is str:
      print('reading 1 file')
      files = [file]
    else:
      print('reading files')
      files = file
  elif folder is not None:
    print("reading from folder")
    files = [i for i in os.listdir(folder) if i.endswith(".xml")]
  else:
    raise ValueError('need folder or file')
  df = pd.DataFrame()
  for file in files:
    res = []
    a = open(file, "r").read()
    xmldict = xmltodict.parse(a)
    data = xmldict['Workbook']['Worksheet']['Table']["Row"]
    for val in data:
      res.append([v['Data']['#text'] if "#text" in v['Data']
                  else None for v in val["Cell"]])
    res = pd.DataFrame(data=res[2:], columns=res[0])
    df = df.append(res)
  if rename is not None:
    df = df.rename(columns=rename)
  return df

def makeCellosaurusExport(ftp = "https://ftp.expasy.org/databases/cellosaurus/cellosaurus.txt",
                          reload=False, dropped=["ACH-002260", "ACH-001741",'ACH-001189',],
                          ancestry = {'African':[], 'Native American':[], 'East Asian, North':[], 
                          'East Asian, South': [], "South Asian":[],  "European, North":[], 
                          'European, South':[]}):
  """
  make a df from cellosaurus' human cancer cell line data

  Args:
    ftp (str): the ftp link to the cellosaurus latest txt db export

  Returns:
    pd.DataFrame: indexed by cellosaurus id and with "depmap_id", "id", "disease", "age", 
    "sex", "patient_id", "parent_id", "date", "synonyms", "has_issues", "comments"
  """
  print('make sure to have wget installed')
  if reload or not os.path.exists('/tmp/cellosaurus.txt'):
    subprocess.run('wget '+ ftp +' -O /tmp/cellosaurus.txt')
  l = open('/tmp/cellosaurus.txt', 'r')
  print("""

    ---------  ---------------------------     ----------------------
    Line code  Content                         Occurrence in an entry
    ---------  ---------------------------     ----------------------
    ID         Identifier (cell line name)     Once; starts an entry
    AC         Accession (CVCL_xxxx)           Once
    AS         Secondary accession number(s)   Optional; once
    SY         Synonyms                        Optional; once
    DR         Cross-references                Optional; once or more
    RX         References identifiers          Optional: once or more
    WW         Web pages                       Optional; once or more
    CC         Comments                        Optional; once or more
    ST         STR profile data                Optional; once or more
    DI         Diseases                        Optional; once or more
    OX         Species of origin               Once or more
    HI         Hierarchy                       Optional; once or more
    OI         Originate from same individual  Optional; once or more
    SX         Sex of cell                     Optional; once
    AG         Age of donor at sampling        Optional; once
    CA         Category                        Once
    DT         Date (entry history)            Once
    //         Terminator                      Once; ends an entry

  """)
  v = []
  CL=False
  Homo=False
  age = "U"
  sex = "U"
  date = "U"
  parent = ""
  ID = ""
  SY = ""
  DI = ""
  atcclink = ""
  dsmzlink = ""
  doublingt = ""
  hasebv = False
  origin = ""
  isMeta = False
  characteristics = ""
  ancestry = ""
  issue = ""
  instability = ""
  transfected = ""
  depmap = set()
  lid = None
  individual= "ID-"+randomString(stringLength=6, stype='all', withdigits=True)
  comments=[""]
  tofind={}
  cl = {}
  while True:
    # Get next line from file
    line = l.readline()
    # if line is empty
    # end of file is reached
    if not line:
      break
    else:
      line = line[:-1]
    if line[:2]=="AC":
      lid = line[5:]
      if lid in cl:
        lid=None
    elif line[:2] =="//":
      depmap-= set(dropped)
      if ((CL and Homo and ID) or depmap) and (lid is not None):
        if len(depmap)>1:
          raise ValueError('multiple depmap mapping to one line', depmap)
        elif len(depmap)==1:
          depmap=depmap.pop()
        else:
          depmap=''
        v.append(depmap)
        v.append(ID)
        v.append(DI)
        v.append(age)
        v.append(sex)
        v.append(individual)
        v.append(parent)
        v.append(date)
        v.append(SY)
        v.append(' | '.join(comments))
        v.append(atcclink)
        v.append(dsmzlink)
        v.append(doublingt)
        v.append(hasebv)
        v.append(origin)
        v.append(isMeta)
        v.append(characteristics)
        v.append(ancestry)
        v.append(issue)
        v.append(instability)
        v.append(transfected)
        cl.update({lid: v})
      v = []
      CL=False
      Homo=False
      age = ""
      sex = ""
      date = ""
      parent = ""
      ID = ""
      SY = ""
      DI = ""
      atcclink = ""
      dsmzlink = ""
      doublingt = ""
      hasebv = False
      origin = ""
      isMeta = False
      characteristics = ""
      ancestry = ""
      issue = ""
      instability = ""
      transfected = ""
      depmap = set()
      lid = None
      individual= "ID-"+randomString(stringLength=6, stype='all', withdigits=True)
      comments=[""]
    elif line[:2] == "HI":
      parent = line[5:].split(' !')[0]
      if parent in cl:
        individual = cl[parent][5]
      else:
        tofind.update({lid:parent})
    elif line[:2] == "OX":
      if "Homo sapiens" in line:
        Homo=True
    elif line[:2] == "OI":
      name = line[5:].split(' !')[0]
      # adding its same id
      if name in cl:
        individual = cl[name][5]
      else:
        tofind.update({lid:name})
    elif line[:2] == "ID":
      ID = line[5:]
    elif line[:2] == "SY":
      SY = line[5:]
    elif line[:2] == "DI":
      DI += line[5:] + "; "
    elif line[:2] == "DT":
      date = line.split("Created: ")[-1].split(";")[0]
    elif line[:2] == "DR":
      if 'DepMap; ' in line:
        depmap.add(line.split('DepMap; ')[-1])
      if "ATCC" in line:
        atcclink = line.split('; ')[-1]
      if "DSMZ" in line:
        dsmzlink = line.split('; ')[-1]
    elif line[:2] == "CA":
      if 'Cancer cell line' in line:
        CL=True
    elif line[:2] == "AG":
      age = line[5:]
      i = age.split('Y')[0]
      if i.endswith('FW'):
        #foetus
        i = "Fetus"
      elif i=="C":
        i="Children"
      elif i in ['Embryonic stage', 'Late embryonic stage', "Blastocyst stage", "Neonate larva"]:
        i='Embryo'
      elif i=="Age unspecified":
        i="U"
      elif i not in ['Adult', "Children", "Fetus", "Embryo"]:
        if i.endswith('M') or i.endswith('W') or i.endswith('D'):
          # 0Y
          i=0
        elif i.endswith('Y'):
          i = i.split('-')[0].split('<')[-1].split('>')[-1]
          i = float(i)
      age = i
    elif line[:2] == "SX":
      sex = line[5:]
      if sex == "":
        sex="U"
      if sex == 'Sex unspecified':
        sex="U"
    elif line[:2] == "CC":
      comments.append(line[5:])
      if 'Problematic cell line:' in line:
        issue += line.split("Problematic cell line: ")[-1] + " "
      if 'Discontinued: DepMap; ' in line:
        issue += "Removed from DepMap; "
        depmap.remove(line.split(';')[1][1:])
      if "Doubling time: " in line:
        doublingt = line.split("Doubling time: ")[-1].split('(')[0]
        if 'everal week' in doublingt:
          t = 13*24
        else:
          t = float(doublingt.split(' ')[0].split('-')[0].split('~')[-1].split('>')[-1].split('=')[-1].split('<')[-1])
          if 'day' in doublingt:
            t = t*24
        doublingt = int(t)
      if "Epstein-Barr virus (EBV)" in line:
        hasebv = True
      if "Derived from sampling site: " in line:
        origin = line.split("Derived from sampling site: ")[-1][:-1]
      if "Derived from metastatic site:" in line:
        origin = line.split("Derived from metastatic site: ")[-1][:-1]
        isMeta = True
      if "Characteristics: " in line:
        characteristics = line.split("Characteristics: ")[-1][:-1]
      if "Genome ancestry: " in line:
        ancestry = line.split("Genome ancestry: ")[-1][:-1]

      if "Caution: " in line:
        issue = line.split("Caution: ")[-1][:-1]
      if "Microsatellite instability: " in line:
        instability = line.split(" (")[1].split(')')[0]
      if "Transfected with: " in line:
        transfected += line.split("Transfected with: ")[-1][:-1] + ' | '
  l.close()
  for k, v in tofind.items():
    if k in cl and v in cl:
      cl[k][5] = cl[v][5]
  cld = pd.DataFrame(data=cl.values(), columns=["depmap_id",
                                      "id",
                                      "disease",
                                      "age",
                                      "sex",
                                      "patient_id",
                                      "parent_id",
                                      "date",
                                      "synonyms",
                                      "comments",
                                      "atcclink",
                                      "dsmzlink",
                                      "doublingt",
                                      "hasebv",
                                      "origin",
                                      "isMeta",
                                      "characteristics",
                                      "ancestry",
                                      "issue",
                                      "instability",
                                      "transfected"], index=cl.keys())
  return cld

def getAncestry(df, col="ancestry", ancestry = {'African':[], 'Native American':[], 'East Asian, North':[],
  'East Asian, South': [], "South Asian":[], "European, North":[], 'European, South':[]}):
  """getAncestry returns a df with the ancestries

  Args:
      df ([type]): [description]
      col (str, optional): [description]. Defaults to "ancestry".
      ancestry (dict, optional): [description]. Defaults to {'African':[], 'Native American':[], 'East Asian, North':[], 'East Asian, South': [], "South Asian":[], "European, North":[], 'European, South':[]}.

  Returns:
      pd.DataFrame: [description]
  """
  name = []
  for i, anc in df[col].iteritems():
    if anc=="":
      continue
    for place in anc.split('; '):
      ori, perc = place.split('%')[0].split('=')
      ancestry[ori].append(float(perc))
    name.append(i)
  return pd.DataFrame(data=ancestry, index=name)
    

def _fetchFromServer(ensemble_server, attributes):
  server = BiomartServer(ensemble_server)
  ensmbl = server.datasets['hsapiens_gene_ensembl']
  res = pd.read_csv(io.StringIO(ensmbl.search({
    'attributes': attributes
  }, header=1).content.decode()), sep='\t')
  return res

def generateGeneNames(ensemble_server="http://nov2020.archive.ensembl.org/biomart", 
                      useCache=False, cache_folder='/'.join(__file__.split('/')[:-3])+"/", attributes=[]):
  """generate a genelist dataframe from ensembl's biomart

  Args:
      ensemble_server ([type], optional): [description]. Defaults to ENSEMBL_SERVER_V.
      useCache (bool, optional): [description]. Defaults to False.
      cache_folder ([type], optional): [description]. Defaults to CACHE_PATH.

  Raises:
      ValueError: [description]

  Returns:
      [type]: [description]
  """
  attr = ['ensembl_gene_id', 'clone_based_ensembl_gene', 'hgnc_symbol', 'gene_biotype',
             'entrezgene_id']
  assert cache_folder[-1] == '/'
  
  cache_folder = os.path.expanduser(cache_folder)
  createFoldersFor(cache_folder)
  cachefile = os.path.join(cache_folder, '.biomart.csv')
  if useCache & os.path.isfile(cachefile):
    print('fetching gene names from biomart cache')
    res = pd.read_csv(cachefile)
  else:
    print('downloading gene names from biomart')
    res = _fetchFromServer(ensemble_server, attr + attributes)
    res.to_csv(cachefile, index=False)

  res.columns = attr+attributes
  if type(res) is not type(pd.DataFrame()):
    raise ValueError('should be a dataframe')
  res = res[~(res[
      'clone_based_ensembl_gene'].isna() & res['hgnc_symbol'].isna())]
  res.loc[res[res.hgnc_symbol.isna()].index, "hgnc_symbol"] = \
      res[res.hgnc_symbol.isna()
          ]['clone_based_ensembl_gene']

  return res


# GCPFunction.py

from google.cloud import storage
import os
import subprocess
import re
from genepy.utils import helper as h
import signal


def list_blobs_with_prefix(bucket_name, prefix, delimiter=None):
    """Lists all the blobs in the bucket that begin with the prefix.

    This can be used to list all blobs in a "folder", e.g. "public/".

    The delimiter argument can be used to restrict the results to only the
    "files" in the given "folder". Without the delimiter, the entire tree under
    the prefix is returned. For example, given these blobs:

            /a/1.txt
            /a/b/2.txt

    If you just specify prefix = '/a', you'll get back:

            /a/1.txt
            /a/b/2.txt

    However, if you specify prefix='/a' and delimiter='/', you'll get back:

            /a/1.txt

    """
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(bucket_name)
    ret = []
    blobs = bucket.list_blobs(prefix=prefix, delimiter=delimiter)
    for blob in blobs:
        ret.append(blob.name)
    return ret


def mvFiles(files, location, group=50, listen_to_errors=False):
    """
    move a set of files in parallel (when the set is huge)

    Args:
    ----
        files: gs paths
        location: to move the files to
        group: files to do in parallel
    """
    by = len(files) if len(files) < group else group
    for sfiles in h.grouped(files, by):
        a = ""
        for val in sfiles:
            a += val + " "
        code = os.system("gsutil -m mv " + a + location)
        if code != 0 and listen_to_errors:
            print("pressed ctrl+c or command failed")
            break


def lsFiles(files, add="", group=50):
    """
    list a set of files in parallel (when the set is huge)

    Args:
    ----
        files: gs paths
        add: additional params to add
        group: files to do in parallel
    """
    print("listing files in gs")
    by = len(files) if len(files) < group else group
    res = []
    for sfiles in h.grouped(files, by):
        a = ""
        for val in sfiles:
            a += val + " "
        data = subprocess.run(
            "gsutil -m ls " + add + " " + a, capture_output=True, shell=True
        )
        if data.returncode != 0:
            if "One or more URLs matched no objects" not in str(data.stderr):
                raise ValueError("issue with the command: " + str(data.stderr))
        if len(str(data.stdout)) < 4:
            return []
        res += (
            str(data.stdout)[2:-1].split("\\n")[:-1]
            if "L" not in add
            else ["gs://" + i for i in str(data.stdout).split("\\ngs://")]
        )
        if "TOTAL:" in res[-1] and "L" not in add:
            res = res[:-1]
    return res


def cpFiles(files, location, group=50):
    """
    copy a set of files in parallel (when the set is huge)

    Args:
    ----
        files: gs paths
        location to copy
        group: files to do in parallel
    """
    by = len(files) if len(files) < group else group
    for sfiles in h.grouped(files, by):
        a = ""
        for val in sfiles:
            a += val + " "
        code = os.system("gsutil -m cp " + a + location)
        if code != 0:
            print("pressed ctrl+c or command failed")
            break


def catFiles(files, group=50, split=False, cut=False):
    """
    copy a set of files in parallel (when the set is huge)

    Args:
    ----
        files: gs paths
        location to copy
        group: files to do in parallel
        cut: split all lines into chunks of size cut
        split: split lines by split e.g. \\n
    """
    by = len(files) if len(files) < group else group
    res = []
    for i, sfiles in enumerate(h.grouped(files, by)):
        print(i / (len(files) / by))
        a = ""
        for val in sfiles:
            a += val + " "
        data = subprocess.run("gsutil -m cat " + a, capture_output=True, shell=True)
        if data.returncode != 0:
            if "One or more URLs matched no objects" not in str(data.stderr):
                print(ValueError("issue with the command: " + str(data.stderr)))
                return res
        if len(str(data.stdout)) < 4:
            return []
        resa = str(data.stdout)[2:-1]
        if cut:
            res += [resa[i * cut : (i + 1) * cut] for i in range(int(len(resa) / cut))]
        elif split:
            res += resa.split(split)
        else:
            res += [resa]
    return res


def rmFiles(files, group=50, add="", dryrun=True):
    """
    remove a set of files in parallel (when the set is huge)

    Args:
    ----
        files: gs paths
        group: number to do in parallel
        add: additional gsutil cp params
    """
    by = len(files) if len(files) < group else group
    for sfiles in h.grouped(files, by):
        a = ""
        for val in sfiles:
            a += " " + val
        if add:
            add = " " + add
        if dryrun:
            print("gsutil -m rm" + add + a)
        else:
            code = os.system("gsutil -m rm" + add + a)
            if code != 0:
                print("pressed ctrl+c or command failed")
                break


def recoverFiles(files, cores=1):
    """
    recover a set of files in parallel that were erased

    files need to have their #id appended found using ls -al file

    Args:
    ----
        files: gs paths
        location: to move the files to
    """
    cmd = ["gsutil mv " + f + " " + f.split("#")[0] for f in files]
    h.parrun(cmd, cores=cores)


def folderRN(gspath, newpath, cores=1):
    """ """
    lis = lsFiles([gspath])
    if lis != 0:
        h.parrun(["gsutil -m mv " + val + " " + newpath for val in lis], cores=cores)
    else:
        raise ValueError("no such folder")


def patternRN(
    rename_dict,
    location,
    wildcards,
    types=[],
    dryrun=True,
    check_dependencies=True,
    cores=1,
):
    """
    rename/move a bunch of GCP objects found in some specific places

    Args:
    -----
        rename_dict: dict(prevName,newName)
        location:
        wildcards: list[str] can be one of  ['**', '.*', '*.','-.*'] if needs to be
                    ** means any occurence of this file in any folder will change its name
                    .* means all file unregarding of the suffix, will rename them all a.bam [a]da.bai to b.bam, [b]da.bai
                    *. means all files with the suffix, will change the suffix of these files from a to b
                    -.* means all file unregarding of the suffix, will rename them. not just replacing the a part with a to b but the full file name [a]dea.bam to b.bam
        types: Nothing yet
        test: if test, just shows the command but does not run it
        cores:  cores tells on how many processor to parallelize the tas#k
    """
    val = []
    for k, v in rename_dict.items():
        val.append(v)
        if k in val and check_dependencies:
            raise ValueError("circular dependency in the rename with key " + k)
    for k, v in rename_dict.items():
        loc = location
        if "**" in wildcards:
            loc += "**/"
        if "*." in wildcards or "-.*" in wildcards:
            loc += "*"
        loc += k
        if ".*" in wildcards or "-.*" in wildcards:
            loc += "*"
        res = os.popen("gsutil -m ls " + loc).read().split("\n")[:-1]
        print("found " + str(len(res)) + " files to rename")
        if "-.*" in wildcards:
            cmd = [
                "gsutil mv "
                + val
                + " "
                + "/".join(val.split("/")[:-1])
                + "/"
                + v
                + "."
                + ".".join(val.split("/")[-1].split(".")[1:])
                for val in res
            ]
        else:
            cmd = ["gsutil mv " + val + " " + val.replace(k, v) for val in res]
        if dryrun:
            print(cmd)
        else:
            h.parrun(cmd, cores=cores)


def get_all_sizes(folder, suffix="*"):
    """
    will sort and list all the files by their sizes.

    If some files have the same size, will list them together

    Args:
    ----
            folder: gs folder path
            suffix: of a specific file type

    Returns:
    -------
            dict(sizes:[paths])
    """
    samples = os.popen("gsutil -m ls -al " + folder + "**." + suffix).read().split("\n")
    # compute size filepath
    sizes = {
        "gs://"
        + val.split("gs://")[1].split("#")[0]: int(
            re.split("\d{4}-\d{2}-\d{2}", val)[0]
        )
        for val in samples[:-2]
    }
    names = {}
    for k, val in sizes.items():
        if val in names:
            names[val].append(k)
        else:
            names[val] = [k]
    if names == {}:
        # we didn't find any valid file paths
        print("We didn't find any valid file paths in folder: " + str(folder))
    return names


def exists(val):
    """
    tells if a gcp path exists
    """
    if type(val) is str:
        return os.popen("gsutil ls " + val).read().split("\n")[0] == val
    elif type(val) is list:
        rest = set(val) - set(lsFiles(val))
        return len(rest) == 0, rest


def extractSize(val):
    """
    extract the size from the string returned by an ls -l|a command
    """
    return "gs://" + val.split("gs://")[1].split("#")[0], int(
        re.split("\d{4}-\d{2}-\d{2}", val)[0]
    )


def extractTime(val):
    """
    extract the size from the string returned by an ls -l|a command
    """
    return val.split("  ")[1].split("T")[0]


def extractPath(val):
    """
    extract the path from the string returned by an ls -l|a command
    """
    return "gs://" + val.split("gs://")[1].split("#")[0]


def extractHash(val, typ="crc32c"):
    """
    extract the crc32 from the string returned by an ls -L command

    Args:
    ----
        type: flag ['crc32c','md5']
    """
    if "    Hash (crc32c):" in val and typ == "crc32c":
        return (
            val.split("    Hash (crc32c):          ")[-1]
            .split("\\\\n")[0]
            .split("\\n")[0]
        )
    elif "    Hash (md5):" in val and typ == "md5":
        return (
            val.split("    Hash (md5):          ")[-1].split("\\\\n")[0].split("\\n")[0]
        )
    else:
        return None


async def shareFiles(flist, users):
    """
    will share a list of files from gcp with a set of users.

    Args:
    ----
      users: list[str] of users' google accounts
      flist: list[str] of google storage path for which you want to share data

    """
    if type(users) is str:
        users = [users]
    for user in users:
        files = ""
        for i in flist:
            files += " " + i
        code = os.system("gsutil -m acl ch -ru " + user + ":R " + files)
        if code == signal.SIGINT:
            print("Awakened")
            break
    print("the files are stored here:\n\n")
    print(flist)
    print("\n\njust install and use gsutil to copy them")
    print("https://cloud.google.com/storage/docs/gsutil_install")
    print("https://cloud.google.com/storage/docs/gsutil/commands/cp")


def deleteOldVersions(path, onlymetagene=None, **kwargs):
    """ """
    data = subprocess.run("gsutil -m ls -alh " + path, capture_output=True, shell=True)
    if data.returncode != 0:
        if "One or more URLs matched no objects" not in str(data.stderr):
            print(ValueError("issue with the command: " + str(data.stderr)))
            return res
    if len(str(data.stdout)) < 4:
        return []
    resa = str(data.stdout)[2:-1].split("\\n")[:-2]
    torm = []
    for i, val in enumerate(resa):
        if onlymetagene is not None:
            if "metageneration=" + str(metagene) in val:
                name = "gs://" + val.split(" gs://")[1].split("  ")[0]
                torm.append(name)
            else:
                prevname = torm[-1].split("#")[0]
                name = "gs://" + val.split(" gs://")[1].split("#")[0]
                if prevname != name:
                    print(prevname + " is unique and won't be deleted")
                    torm.pop()
        else:
            if "metageneration=" + str(1) not in val:
                name = "gs://" + val.split(" gs://")[1].split("  ")[0]
                torm.append(name)
                if (
                    "gs://" + resa[i + 1].split(" gs://")[1].split("#")[0]
                    != name.split("#")[0]
                ):
                    print(name + " is unique and won't be deleted")
                    torm.pop()
    print(h.dups([val.split("#")[0] for val in torm]))

    return rmFiles(torm, **kwargs)
